/*
 * Generic Hash and HMAC Program
 *
 * Copyright (C) 2009 2011 2016 Harald von Fellenberg <hvf@hvf.ch>
 *
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 3 of the License, or 
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

/* 29.09.2024 hvf add kmac128, kmac256, kmacxof128, kmacxof256  */

#include "generic.h"
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "KeccakHash.h"


void print_SHA3_CTX (SHA3_CTX *context);
void print_EXTRA (EXTRA_DATA extra);

/* 
 * parameter safe wrappers for SHA3 routines for each hash length
 */

 /*************************** 224 ************************************/

HashReturn SHA3_224_init (hashState  *state, int hashbitlen)
{
    /* verify correct hash length   */
    if (hashbitlen != HASH_BITLENGTH_SHA3_224)
        return BAD_HASHBITLEN;

    /* allocate context and fill it */
    SHA3_CTX *context = (SHA3_CTX *)malloc (sizeof (SHA3_CTX));
    memset (context, 0, sizeof (SHA3_CTX));
    context->hashbitlen = HASH_BITLENGTH_SHA3_224;
    context->magic = HASH_MAGIC_SHA3_224;
	*state = (hashState *) context;
	return Keccak_HashInitialize_SHA3_224 (context);
}

HashReturn  SHA3_224_update (
    hashState state,          /* previously initialized context */
    const BitSequence *buffer,  /* bit buffer, first bit is MSB in [0] */
    DataLength databitlen)      /* number of bits to process from buffer */
{
    /* can be called once or many times */
    /* verify correct hashbitlen and magic  */

    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_SHA3_224)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_SHA3_224)
        return BAD_ALGORITHM;

	return Keccak_HashUpdate (context, buffer, databitlen);
}

HashReturn  SHA3_224_final (hashState state, BitSequence *hashval)
{
    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_SHA3_224)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_SHA3_224)
        return BAD_ALGORITHM;

	// if (!hashval)
	//	hashval = context->out;

	HashReturn retval = Keccak_HashFinal (context, context->out);
	if (hashval) memcpy (hashval, context->out, HASH_LENGTH_SHA3_224);
	return retval;
}

HashReturn SHA3_224_hash (int hashbitlen, const BitSequence *data,
                      DataLength databitlen, BitSequence *hashval)
{
    hashState   state;
    HashReturn  retval;

    retval = SHA3_224_init (&state, HASH_BITLENGTH_SHA3_224);
    if (retval != SUCCESS) {
        fprintf (stderr, "SHA3_224_init failed, reason %d, hash length %d\n",
                 retval, HASH_BITLENGTH_SHA3_224);
        exit (1);
    }

    retval = SHA3_224_update (state, data, databitlen);
    if (retval != SUCCESS) {
        fprintf (stderr, "SHA3_224_update failed, reason %d\n", retval);
        exit (1);
    }

    retval = SHA3_224_final (state, hashval);
    if (retval != SUCCESS) {
        fprintf (stderr, "SHA3_224_final failed, reason %d\n", retval);
        exit (1);
    }
    free (state);
    return retval;
}

/*
 * three functions in MD5 style for each hash length
 */

/* Digests a file and prints the result.
 */
HashReturn SHA3_224_File (hashState state, FILE *in)
{
	SHA3_CTX *context = (SHA3_CTX *) state;
	int len;
	unsigned char buffer[BUFFERSIZE];
	HashReturn retval;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_SHA3_224)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_SHA3_224)
		return BAD_ALGORITHM;

	while ((len = fread (buffer, 1, BUFFERSIZE, in))) {
		retval = SHA3_224_update (context, buffer, (DataLength)len << 3);
		if (retval != SUCCESS) return retval;
	}
	retval = SHA3_224_final (context, NULL);

	fclose (in);
	return retval;
}

HashReturn SHA3_224_HashToByte (hashState state, BYTE *out) 
{
	SHA3_CTX *context = (SHA3_CTX *) state;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_SHA3_224)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_SHA3_224)
		return BAD_ALGORITHM;

	memcpy (out, context->out, HASH_LENGTH_SHA3_224);
	return SUCCESS;
}


 /*************************** 256 ************************************/

HashReturn SHA3_256_init (hashState  *state, int hashbitlen)
{
    /* verify correct hash length   */
    if (hashbitlen != HASH_BITLENGTH_SHA3_256)
        return BAD_HASHBITLEN;

    /* allocate context and fill it */
    SHA3_CTX *context = (SHA3_CTX *)malloc (sizeof (SHA3_CTX));
    memset (context, 0, sizeof (SHA3_CTX));
    context->hashbitlen = HASH_BITLENGTH_SHA3_256;
    context->magic = HASH_MAGIC_SHA3_256;
	*state = (hashState *) context;
	return Keccak_HashInitialize_SHA3_256 (context);
}

HashReturn  SHA3_256_update (
    hashState state,          /* previously initialized context */
    const BitSequence *buffer,  /* bit buffer, first bit is MSB in [0] */
    DataLength databitlen)      /* number of bits to process from buffer */
{
    /* can be called once or many times */
    /* verify correct hashbitlen and magic  */

    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_SHA3_256)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_SHA3_256)
        return BAD_ALGORITHM;

	return Keccak_HashUpdate (context, buffer, databitlen);
}

HashReturn  SHA3_256_final (hashState state, BitSequence *hashval)
{
    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_SHA3_256)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_SHA3_256)
        return BAD_ALGORITHM;

	//if (!hashval)
		//hashval = context->out;

	//return Keccak_HashFinal (context, hashval);
	HashReturn retval = Keccak_HashFinal (context, context->out);
	if (hashval) memcpy (hashval, context->out, HASH_LENGTH_SHA3_256);
	return retval;
}

HashReturn SHA3_256_hash (int hashbitlen, const BitSequence *data,
                      DataLength databitlen, BitSequence *hashval)
{
    hashState   state;
    HashReturn  retval;

    retval = SHA3_256_init (&state, HASH_BITLENGTH_SHA3_256);
    if (retval != SUCCESS) {
        fprintf (stderr, "SHA3_256_init failed, reason %d, hash length %d\n",
                 retval, HASH_BITLENGTH_SHA3_256);
        exit (1);
    }

    retval = SHA3_256_update (state, data, databitlen);
    if (retval != SUCCESS) {
        fprintf (stderr, "SHA3_256_update failed, reason %d\n", retval);
        exit (1);
    }

    retval = SHA3_256_final (state, hashval);
    if (retval != SUCCESS) {
        fprintf (stderr, "SHA3_256_final failed, reason %d\n", retval);
        exit (1);
    }
    free (state);
    return retval;
}

/*
 * three functions in MD5 style for each hash length
 */

/* Digests a file and prints the result.
 */
HashReturn SHA3_256_File (hashState state, FILE *in)
{
	SHA3_CTX *context = (SHA3_CTX *) state;
	int len;
	unsigned char buffer[BUFFERSIZE];
	HashReturn retval;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_SHA3_256)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_SHA3_256)
		return BAD_ALGORITHM;

	while ((len = fread (buffer, 1, BUFFERSIZE, in))) {
		retval = SHA3_256_update (context, buffer, (DataLength)len << 3);
		if (retval != SUCCESS) return retval;
	}
	retval = SHA3_256_final (context, NULL);

	fclose (in);
	return retval;
}

HashReturn SHA3_256_HashToByte (hashState state, BYTE *out) 
{
	SHA3_CTX *context = (SHA3_CTX *) state;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_SHA3_256)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_SHA3_256)
		return BAD_ALGORITHM;

	memcpy (out, context->out, HASH_LENGTH_SHA3_256);
	return SUCCESS;
}


 /*************************** 384 ************************************/

HashReturn SHA3_384_init (hashState  *state, int hashbitlen)
{
    /* verify correct hash length   */
    if (hashbitlen != HASH_BITLENGTH_SHA3_384)
        return BAD_HASHBITLEN;

    /* allocate context and fill it */
    SHA3_CTX *context = (SHA3_CTX *)malloc (sizeof (SHA3_CTX));
    memset (context, 0, sizeof (SHA3_CTX));
    context->hashbitlen = HASH_BITLENGTH_SHA3_384;
    context->magic = HASH_MAGIC_SHA3_384;
	*state = (hashState *) context;
	return Keccak_HashInitialize_SHA3_384 (context);
}

HashReturn  SHA3_384_update (
    hashState state,          /* previously initialized context */
    const BitSequence *buffer,  /* bit buffer, first bit is MSB in [0] */
    DataLength databitlen)      /* number of bits to process from buffer */
{
    /* can be called once or many times */
    /* verify correct hashbitlen and magic  */

    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_SHA3_384)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_SHA3_384)
        return BAD_ALGORITHM;

	return Keccak_HashUpdate (context, buffer, databitlen);
}

HashReturn  SHA3_384_final (hashState state, BitSequence *hashval)
{
    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_SHA3_384)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_SHA3_384)
        return BAD_ALGORITHM;

//	if (!hashval)
		//hashval = context->out;

	//return Keccak_HashFinal (context, hashval);
	HashReturn retval = Keccak_HashFinal (context, context->out);
	if (hashval) memcpy (hashval, context->out, HASH_LENGTH_SHA3_384);
	return retval;
}

HashReturn SHA3_384_hash (int hashbitlen, const BitSequence *data,
                      DataLength databitlen, BitSequence *hashval)
{
    hashState   state;
    HashReturn  retval;

    retval = SHA3_384_init (&state, HASH_BITLENGTH_SHA3_384);
    if (retval != SUCCESS) {
        fprintf (stderr, "SHA3_384_init failed, reason %d, hash length %d\n",
                 retval, HASH_BITLENGTH_SHA3_384);
        exit (1);
    }

    retval = SHA3_384_update (state, data, databitlen);
    if (retval != SUCCESS) {
        fprintf (stderr, "SHA3_384_update failed, reason %d\n", retval);
        exit (1);
    }

    retval = SHA3_384_final (state, hashval);
    if (retval != SUCCESS) {
        fprintf (stderr, "SHA3_384_final failed, reason %d\n", retval);
        exit (1);
    }
    free (state);
    return retval;
}

/*
 * three functions in MD5 style for each hash length
 */

/* Digests a file and prints the result.
 */
HashReturn SHA3_384_File (hashState state, FILE *in)
{
	SHA3_CTX *context = (SHA3_CTX *) state;
	int len;
	unsigned char buffer[BUFFERSIZE];
	HashReturn retval;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_SHA3_384)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_SHA3_384)
		return BAD_ALGORITHM;

	while ((len = fread (buffer, 1, BUFFERSIZE, in))) {
		retval = SHA3_384_update (context, buffer, (DataLength)len << 3);
		if (retval != SUCCESS) return retval;
	}
	retval = SHA3_384_final (context, NULL);

	fclose (in);
	return retval;
}

HashReturn SHA3_384_HashToByte (hashState state, BYTE *out) 
{
	SHA3_CTX *context = (SHA3_CTX *) state;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_SHA3_384)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_SHA3_384)
		return BAD_ALGORITHM;

	memcpy (out, context->out, HASH_LENGTH_SHA3_384);
	return SUCCESS;
}


 /*************************** 512 ************************************/

HashReturn SHA3_512_init (hashState  *state, int hashbitlen)
{
    /* verify correct hash length   */
    if (hashbitlen != HASH_BITLENGTH_SHA3_512)
        return BAD_HASHBITLEN;

    /* allocate context and fill it */
    SHA3_CTX *context = (SHA3_CTX *)malloc (sizeof (SHA3_CTX));
    memset (context, 0, sizeof (SHA3_CTX));
    context->hashbitlen = HASH_BITLENGTH_SHA3_512;
    context->magic = HASH_MAGIC_SHA3_512;
	*state = (hashState *) context;
	return Keccak_HashInitialize_SHA3_512 (context);
}

HashReturn  SHA3_512_update (
    hashState state,          /* previously initialized context */
    const BitSequence *buffer,  /* bit buffer, first bit is MSB in [0] */
    DataLength databitlen)      /* number of bits to process from buffer */
{
    /* can be called once or many times */
    /* verify correct hashbitlen and magic  */

    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_SHA3_512)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_SHA3_512)
        return BAD_ALGORITHM;

	return Keccak_HashUpdate (context, buffer, databitlen);
}

HashReturn  SHA3_512_final (hashState state, BitSequence *hashval)
{
    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_SHA3_512)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_SHA3_512)
        return BAD_ALGORITHM;

	//if (!hashval)
		//hashval = context->out;

	//return Keccak_HashFinal (context, hashval);
	HashReturn retval = Keccak_HashFinal (context, context->out);
	if (hashval) memcpy (hashval, context->out, HASH_LENGTH_SHA3_512);
	return retval;
}

HashReturn SHA3_512_hash (int hashbitlen, const BitSequence *data,
                      DataLength databitlen, BitSequence *hashval)
{
    hashState   state;
    HashReturn  retval;

    retval = SHA3_512_init (&state, HASH_BITLENGTH_SHA3_512);
    if (retval != SUCCESS) {
        fprintf (stderr, "SHA3_512_init failed, reason %d, hash length %d\n",
                 retval, HASH_BITLENGTH_SHA3_512);
        exit (1);
    }

    retval = SHA3_512_update (state, data, databitlen);
    if (retval != SUCCESS) {
        fprintf (stderr, "SHA3_512_update failed, reason %d\n", retval);
        exit (1);
    }

    retval = SHA3_512_final (state, hashval);
    if (retval != SUCCESS) {
        fprintf (stderr, "SHA3_512_final failed, reason %d\n", retval);
        exit (1);
    }
    free (state);
    return retval;
}

/*
 * three functions in MD5 style for each hash length
 */

/* Digests a file and prints the result.
 */
HashReturn SHA3_512_File (hashState state, FILE *in)
{
	SHA3_CTX *context = (SHA3_CTX *) state;
	int len;
	unsigned char buffer[BUFFERSIZE];
	HashReturn retval;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_SHA3_512)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_SHA3_512)
		return BAD_ALGORITHM;

	while ((len = fread (buffer, 1, BUFFERSIZE, in))) {
		retval = SHA3_512_update (context, buffer, (DataLength)len << 3);
		if (retval != SUCCESS) return retval;
	}
	retval = SHA3_512_final (context, NULL);

	fclose (in);
	return retval;
}

HashReturn SHA3_512_HashToByte (hashState state, BYTE *out) 
{
	SHA3_CTX *context = (SHA3_CTX *) state;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_SHA3_512)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_SHA3_512)
		return BAD_ALGORITHM;

	memcpy (out, context->out, HASH_LENGTH_SHA3_512);
	return SUCCESS;
}


#if 0
// shake128, 256 are eplaced by xofshake128, 256 - hvf 07.11.2024
 /*************************** SHAKE128 ************************************/

HashReturn SHAKE128_init (hashState  *state, int hashbitlen)
{
    /* verify correct hash length   */
    if (hashbitlen != HASH_BITLENGTH_SHAKE128)
        return BAD_HASHBITLEN;

    /* allocate context and fill it */
    SHA3_CTX *context = (SHA3_CTX *)malloc (sizeof (SHA3_CTX));
    memset (context, 0, sizeof (SHA3_CTX));
    context->hashbitlen = HASH_BITLENGTH_SHAKE128;
    context->magic = HASH_MAGIC_SHAKE128;
	*state = (hashState *) context;
	return Keccak_HashInitialize_SHAKE128 (context);
}

HashReturn  SHAKE128_update (
    hashState state,          /* previously initialized context */
    const BitSequence *buffer,  /* bit buffer, first bit is MSB in [0] */
    DataLength databitlen)      /* number of bits to process from buffer */
{
    /* can be called once or many times */
    /* verify correct hashbitlen and magic  */

    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_SHAKE128)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_SHAKE128)
        return BAD_ALGORITHM;

	return Keccak_HashUpdate (context, buffer, databitlen);
}

HashReturn  SHAKE128_final (hashState state, BitSequence *hashval)
{
    SHA3_CTX *context = (SHA3_CTX *) state;

	HashReturn	retval;

    if (context->hashbitlen != HASH_BITLENGTH_SHAKE128)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_SHAKE128)
        return BAD_ALGORITHM;

	Keccak_HashFinal (context, context->out);

	// if hashval is NULL, we allocate a buffer for intermediate storage
	if (!hashval) hashval = malloc (SqueezingOutputLength/8);
	size_t	nchunksize = 100;	// squeeze 64 bytes (only) each time
	//size_t squeezechunksize = 8*nchunksize;
	size_t nbytesdone;
	//int bitslefttodo = SqueezingOutputLength;
	for (nbytesdone = 0; nbytesdone < SqueezingOutputLength/8; 
		nbytesdone += nchunksize) {
		if (nbytesdone + nchunksize > SqueezingOutputLength/8) 
			nchunksize = SqueezingOutputLength/8 - nbytesdone;
		retval = Keccak_HashSqueeze (context, context->out, nchunksize*8);
		if (hashval) memcpy (hashval+nbytesdone, context->out, nchunksize);
	}
	// copy back into context->out
	memcpy (context->out, hashval, SqueezingOutputLength/8);

	return retval;
}

HashReturn SHAKE128_hash (int hashbitlen, const BitSequence *data,
                      DataLength databitlen, BitSequence *hashval)
{
    hashState   state;
    HashReturn  retval;

    retval = SHAKE128_init (&state, HASH_BITLENGTH_SHAKE128);
    if (retval != SUCCESS) {
        fprintf (stderr, "SHAKE128_init failed, reason %d, hash length %d\n",
                 retval, HASH_BITLENGTH_SHAKE128);
        exit (1);
    }

    retval = SHAKE128_update (state, data, databitlen);
    if (retval != SUCCESS) {
        fprintf (stderr, "SHAKE128_update failed, reason %d\n", retval);
        exit (1);
    }

    retval = SHAKE128_final (state, hashval);
    if (retval != SUCCESS) {
        fprintf (stderr, "SHAKE128_final failed, reason %d\n", retval);
        exit (1);
    }
    free (state);
    return retval;
}

/*
 * three functions in MD5 style for each hash length
 */

/* Digests a file and prints the result.
 */
HashReturn SHAKE128_File (hashState state, FILE *in)
{
	SHA3_CTX *context = (SHA3_CTX *) state;
	int len;
	unsigned char buffer[BUFFERSIZE];
	HashReturn retval;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_SHAKE128)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_SHAKE128)
		return BAD_ALGORITHM;

	while ((len = fread (buffer, 1, BUFFERSIZE, in))) {
		retval = SHAKE128_update (context, buffer, (DataLength)len << 3);
		if (retval != SUCCESS) return retval;
	}
	retval = SHAKE128_final (context, NULL);

	fclose (in);
	return retval;
}

HashReturn SHAKE128_HashToByte (hashState state, BYTE *out) 
{
	SHA3_CTX *context = (SHA3_CTX *) state;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_SHAKE128)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_SHAKE128)
		return BAD_ALGORITHM;

	memcpy (out, context->out, HASH_LENGTH_SHAKE128);
	return SUCCESS;
}


 /*************************** SHAKE256 ************************************/

HashReturn SHAKE256_init (hashState  *state, int hashbitlen)
{
    /* verify correct hash length   */
    if (hashbitlen != HASH_BITLENGTH_SHAKE256)
        return BAD_HASHBITLEN;

    /* allocate context and fill it */
    SHA3_CTX *context = (SHA3_CTX *)malloc (sizeof (SHA3_CTX));
    memset (context, 0, sizeof (SHA3_CTX));
    context->hashbitlen = HASH_BITLENGTH_SHAKE256;
    context->magic = HASH_MAGIC_SHAKE256;
	*state = (hashState *) context;
	return Keccak_HashInitialize_SHAKE256 (context);
}

HashReturn  SHAKE256_update (
    hashState state,          /* previously initialized context */
    const BitSequence *buffer,  /* bit buffer, first bit is MSB in [0] */
    DataLength databitlen)      /* number of bits to process from buffer */
{
    /* can be called once or many times */
    /* verify correct hashbitlen and magic  */

    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_SHAKE256)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_SHAKE256)
        return BAD_ALGORITHM;

	return Keccak_HashUpdate (context, buffer, databitlen);
}

HashReturn  SHAKE256_final (hashState state, BitSequence *hashval)
{
    SHA3_CTX *context = (SHA3_CTX *) state;

	HashReturn	retval;

    if (context->hashbitlen != HASH_BITLENGTH_SHAKE256)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_SHAKE256)
        return BAD_ALGORITHM;

	Keccak_HashFinal (context, context->out);

	retval = Keccak_HashSqueeze (context, context->out, SqueezingOutputLength);
	if (hashval) memcpy (hashval, context->out, SqueezingOutputLength/8);
	return retval;
}

HashReturn SHAKE256_hash (int hashbitlen, const BitSequence *data,
                      DataLength databitlen, BitSequence *hashval)
{
    hashState   state;
    HashReturn  retval;

    retval = SHAKE256_init (&state, HASH_BITLENGTH_SHAKE256);
    if (retval != SUCCESS) {
        fprintf (stderr, "SHAKE256_init failed, reason %d, hash length %d\n",
                 retval, HASH_BITLENGTH_SHAKE256);
        exit (1);
    }

    retval = SHAKE256_update (state, data, databitlen);
    if (retval != SUCCESS) {
        fprintf (stderr, "SHAKE256_update failed, reason %d\n", retval);
        exit (1);
    }

    retval = SHAKE256_final (state, hashval);
    if (retval != SUCCESS) {
        fprintf (stderr, "SHAKE256_final failed, reason %d\n", retval);
        exit (1);
    }
    free (state);
    return retval;
}

/*
 * three functions in MD5 style for each hash length
 */

/* Digests a file and prints the result.
 */
HashReturn SHAKE256_File (hashState state, FILE *in)
{
	SHA3_CTX *context = (SHA3_CTX *) state;
	int len;
	unsigned char buffer[BUFFERSIZE];
	HashReturn retval;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_SHAKE256)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_SHAKE256)
		return BAD_ALGORITHM;

	while ((len = fread (buffer, 1, BUFFERSIZE, in))) {
		retval = SHAKE256_update (context, buffer, (DataLength)len << 3);
		if (retval != SUCCESS) return retval;
	}
	retval = SHAKE256_final (context, NULL);

	fclose (in);
	return retval;
}

HashReturn SHAKE256_HashToByte (hashState state, BYTE *out) 
{
	SHA3_CTX *context = (SHA3_CTX *) state;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_SHAKE256)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_SHAKE256)
		return BAD_ALGORITHM;

	memcpy (out, context->out, HASH_LENGTH_SHAKE256);
	return SUCCESS;
}
#endif

 /*************************** XOFSHAKE128 ************************************/

HashReturn XOFSHAKE128_init (hashState  *state, int hashbitlen, EXTRA_DATA extra)
{
    /* verify correct hash length   */
    if (hashbitlen != HASH_BITLENGTH_SHAKE128)
        return BAD_HASHBITLEN;

    /* allocate context and fill it */
    SHA3_CTX *context = (SHA3_CTX *)malloc (sizeof (SHA3_CTX));
    memset (context, 0, sizeof (SHA3_CTX));
    context->hashbitlen = HASH_BITLENGTH_SHAKE128;
    context->magic = HASH_MAGIC_SHAKE128;
	context->xof_OK = 1;
	context->xoflength= XOF_LENGTH_XOFSHAKE128;
	context->xofinfinity = 0;
	// to make life easier for base64, chunksize is a multiple of 12
	context->this_chunk_size = XOF_LENGTH_XOFSHAKE128;
	context->more_size = XOF_LENGTH_XOFSHAKE128;

	/* extra data for XOF */
	if (extra) {
		context->base64flag = extra->base64flag;
		context->xofflag = extra->xofflag;
		context->xoflength= extra->xoflength;
		context->more_size= extra->xoflength;
		context->xofinfinity = (context->xoflength == -1);
		// default length is 512 bytes if xoflength == 0 (unspecified)
		if (extra->xoflength == 0) {    // use default length
			context->xoflength = XOF_DEFAULT_LENGTH_XOFSHAKE128;
			context->more_size = XOF_DEFAULT_LENGTH_XOFSHAKE128;
		}
		// length is shorter than default
		if ((context->xoflength > 0) && (context->xoflength < XOF_LENGTH_XOFSHAKE128)) {
			context->more_size = context->this_chunk_size = context->xoflength;
		}
		// length is infinity
		if (extra->xoflength < 0) {		// -1, infinity
			context->xoflength= XOF_LENGTH_XOFSHAKE128;
			context->xoflength= -1;
			context->more_size = -1;
			context->xofinfinity = 1;
		}

		context->binoutflag = extra->binoutflag;	// not yet used
	}

	*state = (hashState *) context;
	return Keccak_HashInitialize_SHAKE128 (context);

}

HashReturn  XOFSHAKE128_update (
    hashState state,          /* previously initialized context */
    const BitSequence *buffer,  /* bit buffer, first bit is MSB in [0] */
    DataLength databitlen)      /* number of bits to process from buffer */
{
    /* can be called once or many times */
    /* verify correct hashbitlen and magic  */

    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_SHAKE128)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_SHAKE128)
        return BAD_ALGORITHM;

	return Keccak_HashUpdate (context, buffer, databitlen);
}

HashReturn  XOFSHAKE128_final (hashState state, BitSequence *hashval)
{
    SHA3_CTX *context = (SHA3_CTX *) state;

	HashReturn	retval;

    if (context->hashbitlen != HASH_BITLENGTH_SHAKE128)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_SHAKE128)
        return BAD_ALGORITHM;

//fprintf(stderr, "\nxofshake128_final: squeezing? %d\n", context->squeezeflag);
	/* check if already squeezing	*/

	if (!context->squeezeflag) {	// skip if squeezing
		retval = Keccak_HashFinal (context, context->out);
		context->squeezeflag = 1;
	}
//fprintf (stderr, "more_size, this_chunk_size = %lld, %d\n", context->more_size, context->this_chunk_size);
	/* we squeeze 600 bytes (a multiple of 12 bytes) to make base64 
	 * conversion easier
	 * total length in bytes is in context->xoflength, 
	 * remaining length in context->more_size
	 * -1 means infinity	*/
    /* 600 bytes is a real hack: it is just a bit longer than the standard
     * length of 512 bytes for shake128, therefore we can use the
     * usual testsuite without stitching the output together
     */

	retval = Keccak_HashSqueeze (context, context->out, XOF_BITLENGTH_XOFSHAKE128);
	// if (hashval) memcpy (hashval, context->out, SqueezingOutputLength/8);
	if (hashval) memcpy (hashval, context->out, context->this_chunk_size);
	if (!context->xofinfinity) {
		// not last block of data
		if (context->more_size > context->this_chunk_size) {
			context->more_size -= context->this_chunk_size;
		} else {	// last block of data
			context->this_chunk_size = context->more_size;
			context->more_size = 0;
		}
	}
	// infinity means: more_size remains on -1, 
	// this_chunk_size remains on its standard value XOF_LENGTH_XOFSHAKE128
//fprintf (stderr, "\nxofshake128_final returns %d\n", retval);
	return retval;
}


/* Digests a file and prints the result.
 */
HashReturn XOFSHAKE128_File (hashState state, FILE *in)
{
	SHA3_CTX *context = (SHA3_CTX *) state;
	int len;
	unsigned char buffer[BUFFERSIZE];
	HashReturn retval;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_SHAKE128)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_SHAKE128)
		return BAD_ALGORITHM;

	while ((len = fread (buffer, 1, BUFFERSIZE, in))) {
		retval = XOFSHAKE128_update (context, buffer, (DataLength)len << 3);
		if (retval != SUCCESS) return retval;
	}
	retval = XOFSHAKE128_final (context, NULL);
	fclose (in);
	return retval;
}

 /*************************** XOFSHAKE256 ************************************/

HashReturn XOFSHAKE256_init (hashState  *state, int hashbitlen, EXTRA_DATA extra)
{
    /* verify correct hash length   */
    if (hashbitlen != HASH_BITLENGTH_SHAKE256)
        return BAD_HASHBITLEN;

    /* allocate context and fill it */
    SHA3_CTX *context = (SHA3_CTX *)malloc (sizeof (SHA3_CTX));
    memset (context, 0, sizeof (SHA3_CTX));
    context->hashbitlen = HASH_BITLENGTH_SHAKE256;
    context->magic = HASH_MAGIC_SHAKE256;
	context->xof_OK = 1;
	context->xoflength= XOF_LENGTH_XOFSHAKE256;
	context->xofinfinity = 0;
	// to make life easier for base64, chunksize is a multiple of 12
	context->this_chunk_size = XOF_LENGTH_XOFSHAKE256;
	context->more_size = XOF_LENGTH_XOFSHAKE256;

	/* extra data for XOF */
	if (extra) {
		context->base64flag = extra->base64flag;
		context->xofflag = extra->xofflag;
		context->xoflength= extra->xoflength;
		context->more_size= extra->xoflength;
		context->xofinfinity = (context->xoflength == -1);
		// default length is 512 bytes if xoflength == 0 (unspecified)
		if (extra->xoflength == 0) {    // use default length
			context->xoflength = XOF_DEFAULT_LENGTH_XOFSHAKE256;
			context->more_size = XOF_DEFAULT_LENGTH_XOFSHAKE256;
		}
		// length is shorter than default
		if ((context->xoflength > 0) && (context->xoflength < XOF_LENGTH_XOFSHAKE256)) {
			context->more_size = context->this_chunk_size = context->xoflength;
		}
		// length is infinity
		if (extra->xoflength < 0) {		// -1, infinity
			context->xoflength= XOF_LENGTH_XOFSHAKE256;
			context->xoflength= -1;
			context->more_size = -1;
			context->xofinfinity = 1;
		}

		context->binoutflag = extra->binoutflag;	// not yet used
	}

	*state = (hashState *) context;
	return Keccak_HashInitialize_SHAKE256 (context);

}

HashReturn  XOFSHAKE256_update (
    hashState state,          /* previously initialized context */
    const BitSequence *buffer,  /* bit buffer, first bit is MSB in [0] */
    DataLength databitlen)      /* number of bits to process from buffer */
{
    /* can be called once or many times */
    /* verify correct hashbitlen and magic  */

    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_SHAKE256)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_SHAKE256)
        return BAD_ALGORITHM;

	return Keccak_HashUpdate (context, buffer, databitlen);
}

HashReturn  XOFSHAKE256_final (hashState state, BitSequence *hashval)
{
    SHA3_CTX *context = (SHA3_CTX *) state;

	HashReturn	retval;

    if (context->hashbitlen != HASH_BITLENGTH_SHAKE256)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_SHAKE256)
        return BAD_ALGORITHM;

	/* check if already squeezing	*/

	if (!context->squeezeflag) {
		retval = Keccak_HashFinal (context, context->out);
		context->squeezeflag = 1;
	}

	/* SqueezingOutputLength is 4096 bits, equal to HASH_BITLENGTH_SHAKE{128,256}
	 * HASH_LENGTH_SHAKE{128,256} is 512 bytes
	 * we squeeze a multiple of 12 bytes (96 bits) to make base64 conversion easier
	 * total length in bytes is in context->xoflength
	 * -1 means infinity	*/

	retval = Keccak_HashSqueeze (context, context->out, XOF_BITLENGTH_XOFSHAKE256);
	// if (hashval) memcpy (hashval, context->out, SqueezingOutputLength/8);
	if (hashval) memcpy (hashval, context->out, context->this_chunk_size);
	if (!context->xofinfinity) {
		if (context->more_size > context->this_chunk_size) {
			context->more_size -= context->this_chunk_size;
		} else {
			context->this_chunk_size = context->more_size;
			context->more_size = 0;
		}
	}

	return retval;
}


/* Digests a file and prints the result.
 */
HashReturn XOFSHAKE256_File (hashState state, FILE *in)
{
	SHA3_CTX *context = (SHA3_CTX *) state;
	int len;
	unsigned char buffer[BUFFERSIZE];
	HashReturn retval;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_SHAKE256)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_SHAKE256)
		return BAD_ALGORITHM;

	while ((len = fread (buffer, 1, BUFFERSIZE, in))) {
		retval = XOFSHAKE256_update (context, buffer, (DataLength)len << 3);
		if (retval != SUCCESS) return retval;
	}
	retval = XOFSHAKE256_final (context, NULL);

	fclose (in);
	return retval;
}

HashReturn XOFSHAKE128_HashToByte (hashState state, BYTE *out) 
{
	SHA3_CTX *context = (SHA3_CTX *) state;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_SHAKE128)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_SHAKE128)
		return BAD_ALGORITHM;

	/* requested length is context->xoflength, bounded */
	size_t len = context->xoflength;
	if (len <= 0) len = XOF_LENGTH_XOFSHAKE128;
	if (len > XOF_LENGTH_XOFSHAKE128) len = XOF_LENGTH_XOFSHAKE128;

	// memcpy (out, context->out, HASH_LENGTH_SHAKE128);
	memcpy (out, context->out, len);
	return SUCCESS;
}


HashReturn XOFSHAKE256_HashToByte (hashState state, BYTE *out) 
{
	SHA3_CTX *context = (SHA3_CTX *) state;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_SHAKE256)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_SHAKE256)
		return BAD_ALGORITHM;

	/* requested length is context->xoflength, bounded */
	size_t len = context->xoflength;
	if (len <= 0) len = XOF_LENGTH_XOFSHAKE256;
	if (len > XOF_LENGTH_XOFSHAKE256) len = XOF_LENGTH_XOFSHAKE256;

	// memcpy (out, context->out, HASH_LENGTH_SHAKE256);
	memcpy (out, context->out, len);
	return SUCCESS;
}


/* 
 * parameter safe wrappers for KECCAK routines for each hash length
 */

 /*************************** 224 ************************************/

HashReturn KECCAK224_init (hashState  *state, int hashbitlen)
{
    /* verify correct hash length   */
    if (hashbitlen != HASH_BITLENGTH_KECCAK_224)
        return BAD_HASHBITLEN;

    /* allocate context and fill it */
    SHA3_CTX *context = (SHA3_CTX *)malloc (sizeof (SHA3_CTX));
    memset (context, 0, sizeof (SHA3_CTX));
    context->hashbitlen = HASH_BITLENGTH_KECCAK_224;
    context->magic = HASH_MAGIC_KECCAK_224;
	*state = (hashState *) context;
	return Keccak_HashInitialize_keccak224 (context);
}

HashReturn  KECCAK224_update (
    hashState state,          /* previously initialized context */
    const BitSequence *buffer,  /* bit buffer, first bit is MSB in [0] */
    DataLength databitlen)      /* number of bits to process from buffer */
{
    /* can be called once or many times */
    /* verify correct hashbitlen and magic  */

    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_KECCAK_224)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_KECCAK_224)
        return BAD_ALGORITHM;

	return Keccak_HashUpdate (context, buffer, databitlen);
}

HashReturn  KECCAK224_final (hashState state, BitSequence *hashval)
{
    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_KECCAK_224)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_KECCAK_224)
        return BAD_ALGORITHM;

	HashReturn retval = Keccak_HashFinal (context, context->out);
	if (hashval) memcpy (hashval, context->out, HASH_LENGTH_KECCAK_224);
	return retval;
}

HashReturn KECCAK224_hash (int hashbitlen, const BitSequence *data,
                      DataLength databitlen, BitSequence *hashval)
{
    hashState   state;
    HashReturn  retval;

    retval = KECCAK224_init (&state, HASH_BITLENGTH_KECCAK_224);
    if (retval != SUCCESS) {
        fprintf (stderr, "KECCAK224_init failed, reason %d, hash length %d\n",
                 retval, HASH_BITLENGTH_KECCAK_224);
        exit (1);
    }

    retval = KECCAK224_update (state, data, databitlen);
    if (retval != SUCCESS) {
        fprintf (stderr, "KECCAK224_update failed, reason %d\n", retval);
        exit (1);
    }

    retval = KECCAK224_final (state, hashval);
    if (retval != SUCCESS) {
        fprintf (stderr, "KECCAK224_final failed, reason %d\n", retval);
        exit (1);
    }
    free (state);
    return retval;
}

/*
 * three functions in MD5 style for each hash length
 */

/* Digests a file and prints the result.
 */
HashReturn KECCAK224_File (hashState state, FILE *in)
{
	SHA3_CTX *context = (SHA3_CTX *) state;
	int len;
	unsigned char buffer[BUFFERSIZE];
	HashReturn retval;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_KECCAK_224)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_KECCAK_224)
		return BAD_ALGORITHM;

	while ((len = fread (buffer, 1, BUFFERSIZE, in))) {
		retval = KECCAK224_update (context, buffer, (DataLength)len << 3);
		if (retval != SUCCESS) return retval;
	}
	retval = KECCAK224_final (context, NULL);

	fclose (in);
	return retval;
}

HashReturn KECCAK224_HashToByte (hashState state, BYTE *out) 
{
	SHA3_CTX *context = (SHA3_CTX *) state;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_KECCAK_224)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_KECCAK_224)
		return BAD_ALGORITHM;

	memcpy (out, context->out, HASH_LENGTH_KECCAK_224);
	return SUCCESS;
}


 /*************************** 256 ************************************/

HashReturn KECCAK256_init (hashState  *state, int hashbitlen)
{
    /* verify correct hash length   */
    if (hashbitlen != HASH_BITLENGTH_KECCAK_256)
        return BAD_HASHBITLEN;

    /* allocate context and fill it */
    SHA3_CTX *context = (SHA3_CTX *)malloc (sizeof (SHA3_CTX));
    memset (context, 0, sizeof (SHA3_CTX));
    context->hashbitlen = HASH_BITLENGTH_KECCAK_256;
    context->magic = HASH_MAGIC_KECCAK_256;
	*state = (hashState *) context;
	return Keccak_HashInitialize_keccak256 (context);
}

HashReturn  KECCAK256_update (
    hashState state,          /* previously initialized context */
    const BitSequence *buffer,  /* bit buffer, first bit is MSB in [0] */
    DataLength databitlen)      /* number of bits to process from buffer */
{
    /* can be called once or many times */
    /* verify correct hashbitlen and magic  */

    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_KECCAK_256)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_KECCAK_256)
        return BAD_ALGORITHM;

	return Keccak_HashUpdate (context, buffer, databitlen);
}

HashReturn  KECCAK256_final (hashState state, BitSequence *hashval)
{
    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_KECCAK_256)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_KECCAK_256)
        return BAD_ALGORITHM;

	HashReturn retval = Keccak_HashFinal (context, context->out);
	if (hashval) memcpy (hashval, context->out, HASH_LENGTH_KECCAK_256);
	return retval;
}

HashReturn KECCAK256_hash (int hashbitlen, const BitSequence *data,
                      DataLength databitlen, BitSequence *hashval)
{
    hashState   state;
    HashReturn  retval;

    retval = KECCAK256_init (&state, HASH_BITLENGTH_KECCAK_256);
    if (retval != SUCCESS) {
        fprintf (stderr, "KECCAK256_init failed, reason %d, hash length %d\n",
                 retval, HASH_BITLENGTH_KECCAK_256);
        exit (1);
    }

    retval = KECCAK256_update (state, data, databitlen);
    if (retval != SUCCESS) {
        fprintf (stderr, "KECCAK256_update failed, reason %d\n", retval);
        exit (1);
    }

    retval = KECCAK256_final (state, hashval);
    if (retval != SUCCESS) {
        fprintf (stderr, "KECCAK256_final failed, reason %d\n", retval);
        exit (1);
    }
    free (state);
    return retval;
}

/*
 * three functions in MD5 style for each hash length
 */

/* Digests a file and prints the result.
 */
HashReturn KECCAK256_File (hashState state, FILE *in)
{
	SHA3_CTX *context = (SHA3_CTX *) state;
	int len;
	unsigned char buffer[BUFFERSIZE];
	HashReturn retval;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_KECCAK_256)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_KECCAK_256)
		return BAD_ALGORITHM;

	while ((len = fread (buffer, 1, BUFFERSIZE, in))) {
		retval = KECCAK256_update (context, buffer, (DataLength)len << 3);
		if (retval != SUCCESS) return retval;
	}
	retval = KECCAK256_final (context, NULL);

	fclose (in);
	return retval;
}

HashReturn KECCAK256_HashToByte (hashState state, BYTE *out) 
{
	SHA3_CTX *context = (SHA3_CTX *) state;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_KECCAK_256)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_KECCAK_256)
		return BAD_ALGORITHM;

	memcpy (out, context->out, HASH_LENGTH_KECCAK_256);
	return SUCCESS;
}


 /*************************** 384 ************************************/

HashReturn KECCAK384_init (hashState  *state, int hashbitlen)
{
    /* verify correct hash length   */
    if (hashbitlen != HASH_BITLENGTH_KECCAK_384)
        return BAD_HASHBITLEN;

    /* allocate context and fill it */
    SHA3_CTX *context = (SHA3_CTX *)malloc (sizeof (SHA3_CTX));
    memset (context, 0, sizeof (SHA3_CTX));
    context->hashbitlen = HASH_BITLENGTH_KECCAK_384;
    context->magic = HASH_MAGIC_KECCAK_384;
	*state = (hashState *) context;
	return Keccak_HashInitialize_keccak384 (context);
}

HashReturn  KECCAK384_update (
    hashState state,          /* previously initialized context */
    const BitSequence *buffer,  /* bit buffer, first bit is MSB in [0] */
    DataLength databitlen)      /* number of bits to process from buffer */
{
    /* can be called once or many times */
    /* verify correct hashbitlen and magic  */

    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_KECCAK_384)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_KECCAK_384)
        return BAD_ALGORITHM;

	return Keccak_HashUpdate (context, buffer, databitlen);
}

HashReturn  KECCAK384_final (hashState state, BitSequence *hashval)
{
    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_KECCAK_384)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_KECCAK_384)
        return BAD_ALGORITHM;

	HashReturn retval = Keccak_HashFinal (context, context->out);
	if (hashval) memcpy (hashval, context->out, HASH_LENGTH_KECCAK_384);
	return retval;
}

HashReturn KECCAK384_hash (int hashbitlen, const BitSequence *data,
                      DataLength databitlen, BitSequence *hashval)
{
    hashState   state;
    HashReturn  retval;

    retval = KECCAK384_init (&state, HASH_BITLENGTH_KECCAK_384);
    if (retval != SUCCESS) {
        fprintf (stderr, "KECCAK384_init failed, reason %d, hash length %d\n",
                 retval, HASH_BITLENGTH_KECCAK_384);
        exit (1);
    }

    retval = KECCAK384_update (state, data, databitlen);
    if (retval != SUCCESS) {
        fprintf (stderr, "KECCAK384_update failed, reason %d\n", retval);
        exit (1);
    }

    retval = KECCAK384_final (state, hashval);
    if (retval != SUCCESS) {
        fprintf (stderr, "KECCAK384_final failed, reason %d\n", retval);
        exit (1);
    }
    free (state);
    return retval;
}

/*
 * three functions in MD5 style for each hash length
 */

/* Digests a file and prints the result.
 */
HashReturn KECCAK384_File (hashState state, FILE *in)
{
	SHA3_CTX *context = (SHA3_CTX *) state;
	int len;
	unsigned char buffer[BUFFERSIZE];
	HashReturn retval;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_KECCAK_384)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_KECCAK_384)
		return BAD_ALGORITHM;

	while ((len = fread (buffer, 1, BUFFERSIZE, in))) {
		retval = KECCAK384_update (context, buffer, (DataLength)len << 3);
		if (retval != SUCCESS) return retval;
	}
	retval = KECCAK384_final (context, NULL);

	fclose (in);
	return retval;
}

HashReturn KECCAK384_HashToByte (hashState state, BYTE *out) 
{
	SHA3_CTX *context = (SHA3_CTX *) state;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_KECCAK_384)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_KECCAK_384)
		return BAD_ALGORITHM;

	memcpy (out, context->out, HASH_LENGTH_KECCAK_384);
	return SUCCESS;
}


 /*************************** 512 ************************************/

HashReturn KECCAK512_init (hashState  *state, int hashbitlen)
{
    /* verify correct hash length   */
    if (hashbitlen != HASH_BITLENGTH_KECCAK_512)
        return BAD_HASHBITLEN;

    /* allocate context and fill it */
    SHA3_CTX *context = (SHA3_CTX *)malloc (sizeof (SHA3_CTX));
    memset (context, 0, sizeof (SHA3_CTX));
    context->hashbitlen = HASH_BITLENGTH_KECCAK_512;
    context->magic = HASH_MAGIC_KECCAK_512;
	*state = (hashState *) context;
	return Keccak_HashInitialize_keccak512 (context);
}

HashReturn  KECCAK512_update (
    hashState state,          /* previously initialized context */
    const BitSequence *buffer,  /* bit buffer, first bit is MSB in [0] */
    DataLength databitlen)      /* number of bits to process from buffer */
{
    /* can be called once or many times */
    /* verify correct hashbitlen and magic  */

    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_KECCAK_512)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_KECCAK_512)
        return BAD_ALGORITHM;

	return Keccak_HashUpdate (context, buffer, databitlen);
}

HashReturn  KECCAK512_final (hashState state, BitSequence *hashval)
{
    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_KECCAK_512)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_KECCAK_512)
        return BAD_ALGORITHM;

	HashReturn retval = Keccak_HashFinal (context, context->out);
	if (hashval) memcpy (hashval, context->out, HASH_LENGTH_KECCAK_512);
	return retval;
}

HashReturn KECCAK512_hash (int hashbitlen, const BitSequence *data,
                      DataLength databitlen, BitSequence *hashval)
{
    hashState   state;
    HashReturn  retval;

    retval = KECCAK512_init (&state, HASH_BITLENGTH_KECCAK_512);
    if (retval != SUCCESS) {
        fprintf (stderr, "KECCAK512_init failed, reason %d, hash length %d\n",
                 retval, HASH_BITLENGTH_KECCAK_512);
        exit (1);
    }

    retval = KECCAK512_update (state, data, databitlen);
    if (retval != SUCCESS) {
        fprintf (stderr, "KECCAK512_update failed, reason %d\n", retval);
        exit (1);
    }

    retval = KECCAK512_final (state, hashval);
    if (retval != SUCCESS) {
        fprintf (stderr, "KECCAK512_final failed, reason %d\n", retval);
        exit (1);
    }
    free (state);
    return retval;
}

/*
 * three functions in MD5 style for each hash length
 */

/* Digests a file and prints the result.
 */
HashReturn KECCAK512_File (hashState state, FILE *in)
{
	SHA3_CTX *context = (SHA3_CTX *) state;
	int len;
	unsigned char buffer[BUFFERSIZE];
	HashReturn retval;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_KECCAK_512)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_KECCAK_512)
		return BAD_ALGORITHM;

	while ((len = fread (buffer, 1, BUFFERSIZE, in))) {
		retval = KECCAK512_update (context, buffer, (DataLength)len << 3);
		if (retval != SUCCESS) return retval;
	}
	retval = KECCAK512_final (context, NULL);

	fclose (in);
	return retval;
}

HashReturn KECCAK512_HashToByte (hashState state, BYTE *out) 
{
	SHA3_CTX *context = (SHA3_CTX *) state;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_KECCAK_512)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_KECCAK_512)
		return BAD_ALGORITHM;

	memcpy (out, context->out, HASH_LENGTH_KECCAK_512);
	return SUCCESS;
}

/**************************** KMAC128 ***************************/
HashReturn myKMAC128_init (hashState  *state, int hashbitlen, EXTRA_DATA extra)
{
    /* verify correct hash length   */
    if (hashbitlen != HASH_BITLENGTH_KMAC_128)
        return BAD_HASHBITLEN;

    const       BitSequence *key    = extra->key;
    BitLength   keyBitLen           = extra->KeyBitLen;
    // BitLength   outputBitLen        = extra->outputBitLen;
    BitLength   outputBitLen        = HASH_BITLENGTH_KMAC_128;
    const BitSequence *customization = extra->customize;
    BitLength   customizeBitLen     = extra->customizeBitLen;
    
    /* allocate context and fill it */
    SHA3_CTX *context   = (SHA3_CTX *)malloc (sizeof (SHA3_CTX));
    memset (context, 0, sizeof (SHA3_CTX));
    context->hashbitlen = HASH_BITLENGTH_KMAC_128;
    context->magic      = HASH_MAGIC_KMAC_128;
    KMAC_Instance *km   = &context->km;
    
	*state = (hashState *) context;

    int retval =  KMAC128_Initialize(km, key, keyBitLen, outputBitLen, 
                    customization, customizeBitLen);
    // printf ("KMAC128_Initialize returns %d\n", retval);
    return retval;

}

HashReturn  myKMAC128_update (
    hashState state,          /* previously initialized context */
    const BitSequence *buffer,  /* bit buffer, first bit is MSB in [0] */
    DataLength databitlen)      /* number of bits to process from buffer */
{
    /* can be called once or many times */
    /* verify correct hashbitlen and magic  */

    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_KMAC_128)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_KMAC_128)
        return BAD_ALGORITHM;

    KMAC_Instance *km   = &context->km;
	int retval = KMAC128_Update (km, buffer, databitlen);
    // printf ("KMAC128_Update returns %d\n", retval);
    return retval;
}

HashReturn  myKMAC128_final (hashState state, BitSequence *hashval)
{
    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_KMAC_128)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_KMAC_128)
        return BAD_ALGORITHM;

    KMAC_Instance *km   = &context->km;
	// if (!hashval)
	//	hashval = context->out;

	HashReturn retval = KMAC128_Final (km, context->out);
    // printf ("KMAC128_Final returns %d\n", retval);
	if (hashval) memcpy (hashval, context->out, HASH_LENGTH_KMAC_128);
	return retval;
}

/* Digests a file and prints the result.
 */
HashReturn myKMAC128_File (hashState state, FILE *in)
{
	SHA3_CTX *context = (SHA3_CTX *) state;
	int len;
	unsigned char buffer[BUFFERSIZE];
	HashReturn retval;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_KMAC_128)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_KMAC_128)
		return BAD_ALGORITHM;

    KMAC_Instance *km   = &context->km;

	while ((len = fread (buffer, 1, BUFFERSIZE, in))) {
		retval = KMAC128_Update (km, buffer, (DataLength)len << 3);
		if (retval != SUCCESS) return retval;
	}
	retval = KMAC128_Final (km, context->out);
    
	fclose (in);
	return retval;
}

HashReturn myKMAC128_HashToByte (hashState state, BYTE *out) 
{
	SHA3_CTX *context = (SHA3_CTX *) state;

	/* verify correct hashbitlen and magic	*/
    if (context->hashbitlen != HASH_BITLENGTH_KMAC_128)
        return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_KMAC_128)
		return BAD_ALGORITHM;

	memcpy (out, context->out, HASH_LENGTH_KMAC_128);
	return SUCCESS;
}

#if 0
HashReturn myKMAC128_hash (int hashbitlen, const BitSequence *data,
                      DataLength databitlen, BitSequence *hashval)
{
    //hashState   state;
    HashReturn  retval;
    KMAC_Instance   km;
    retval =  KMAC128_Initialize(&km, keystring, keylength, 
                HASH_BITLENGTH_KMAC_128, customstring, customBitLen);
    
    if (retval != SUCCESS) {
        fprintf (stderr, "myKMAC128_init failed, reason %d, hash length %d\n",
                 retval, HASH_BITLENGTH_KMAC_128);
        exit (1);
    }

    retval = KMAC128_Update (&km, data, databitlen);
    if (retval != SUCCESS) {
        fprintf (stderr, "myKMAC128_update failed, reason %d\n", retval);
        exit (1);
    }

    /* result is returned in hashval    */
    retval = KMAC128_Final (&km, hashval);
    if (retval != SUCCESS) {
        fprintf (stderr, "myKMAC128_final failed, reason %d\n", retval);
        exit (1);
    }
    //free (state);
    return retval;
}
#endif

/**************************** KMAC256 ***************************/
HashReturn myKMAC256_init (hashState  *state, int hashbitlen, EXTRA_DATA extra)
{
    /* verify correct hash length   */
    if (hashbitlen != HASH_BITLENGTH_KMAC_256)
        return BAD_HASHBITLEN;

    const       BitSequence *key    = extra->key;
    BitLength   keyBitLen           = extra->KeyBitLen;
    BitLength   outputBitLen        = HASH_BITLENGTH_KMAC_256;
    const BitSequence *customization = extra->customize;
    BitLength   customizeBitLen     = extra->customizeBitLen;
    
    /* allocate context and fill it */
    SHA3_CTX *context   = (SHA3_CTX *)malloc (sizeof (SHA3_CTX));
    memset (context, 0, sizeof (SHA3_CTX));
    context->hashbitlen = HASH_BITLENGTH_KMAC_256;
    context->magic      = HASH_MAGIC_KMAC_256;
    KMAC_Instance *km   = &context->km;
    
	*state = (hashState *) context;

    int retval =  KMAC256_Initialize(km, key, keyBitLen, outputBitLen, 
                    customization, customizeBitLen);

    return retval;

}

HashReturn  myKMAC256_update (
    hashState state,          /* previously initialized context */
    const BitSequence *buffer,  /* bit buffer, first bit is MSB in [0] */
    DataLength databitlen)      /* number of bits to process from buffer */
{
    /* can be called once or many times */
    /* verify correct hashbitlen and magic  */

    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_KMAC_256)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_KMAC_256)
        return BAD_ALGORITHM;

	int retval = KMAC256_Update (&context->km, buffer, databitlen);
    return retval;
}

HashReturn  myKMAC256_final (hashState state, BitSequence *hashval)
{
    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_KMAC_256)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_KMAC_256)
        return BAD_ALGORITHM;

	// if (!hashval)
	//	hashval = context->out;

	HashReturn retval = KMAC256_Final (&context->km, context->out);
	if (hashval) memcpy (hashval, context->out, HASH_LENGTH_KMAC_256);
	return retval;
}

/* Digests a file and prints the result.
 */
HashReturn myKMAC256_File (hashState state, FILE *in)
{
	SHA3_CTX *context = (SHA3_CTX *) state;
	int len;
	unsigned char buffer[BUFFERSIZE];
	HashReturn retval;

	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_KMAC_256)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_KMAC_256)
		return BAD_ALGORITHM;

    KMAC_Instance *km   = &context->km;

	while ((len = fread (buffer, 1, BUFFERSIZE, in))) {
		retval = KMAC256_Update (km, buffer, (DataLength)len << 3);
		if (retval != SUCCESS) return retval;
	}
	retval = KMAC256_Final (km, context->out);

	fclose (in);
	return retval;
}
HashReturn myKMAC256_HashToByte (hashState state, BYTE *out) 
{
	SHA3_CTX *context = (SHA3_CTX *) state;

	/* verify correct hashbitlen and magic	*/
    if (context->hashbitlen != HASH_BITLENGTH_KMAC_256)
        return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_KMAC_256)
		return BAD_ALGORITHM;

	memcpy (out, context->out, HASH_LENGTH_KMAC_256);
	return SUCCESS;
}

#if 0
HashReturn myKMAC256_hash (int hashbitlen, const BitSequence *data,
                      DataLength databitlen, BitSequence *hashval)
{
    HashReturn  retval;
    KMAC_Instance   km;
    retval =  KMAC256_Initialize(&km, keystring, keylength, 
                HASH_BITLENGTH_KMAC_256, customstring, customBitLen);
    
    if (retval != SUCCESS) {
        fprintf (stderr, "myKMAC256_init failed, reason %d, hash length %d\n",
                 retval, HASH_BITLENGTH_KMAC_256);
        exit (1);
    }

    retval = KMAC256_Update (&km, data, databitlen);
    if (retval != SUCCESS) {
        fprintf (stderr, "myKMAC256_update failed, reason %d\n", retval);
        exit (1);
    }

    /* result is returned in hashval    */
    retval = KMAC256_Final (&km, hashval);
    if (retval != SUCCESS) {
        fprintf (stderr, "myKMAC256_final failed, reason %d\n", retval);
        exit (1);
    }
    return retval;
}
#endif

void print_SHA3_CTX (SHA3_CTX *context) 
{
    printf("Context\n");
    printf("hashbitlen = %d\n", context->hashbitlen);
    printf("magic = %llx\n", context->magic);
    printf("Keccak sponge length = %ld\n", sizeof (context->sponge)),
    printf("fixedOutputLength = %d\n", context->fixedOutputLength);
    printf("delimitedSuffix = %2x\n", context->delimitedSuffix);
    printf("hash output buffer length = %d\n", SqueezingOutputLength/8);
    printf("hash output buffer length = %ld\n", sizeof(context->out));
    printf("base64flag = %d\n", context->base64flag);
    printf("xof_OK, xofflag = %d, %d\n", context->xof_OK, context->xofflag);
    printf("xoflength, xofinfinity = %lld, %d\n", context->xoflength,
            context->xofinfinity);
    printf("squeezeflag = %d\n", context->squeezeflag);
    printf("this_chunk_size, more_size = %d, %lld\n", context->this_chunk_size,
            context->more_size);
    printf("KMAC_Instance size = %ld\n", sizeof(KMAC_Instance));
    printf ("  cSHAKE_Instance size = %ld\n", sizeof(cSHAKE_Instance));
    printf ("    KeccakWidth1600_SpongeInstance size = %ld\n", sizeof(KeccakWidth1600_SpongeInstance));
    printf("KMAC_Instance:\n"); 
    KMAC_Instance *km   = &context->km;
    cSHAKE_Instance csi = km->csi;
    printf("  cSHAKE_Instance:\n"); 
    KeccakWidth1600_SpongeInstance  sponge;
    printf("    sponge = \n"); 
    printf("      sponge.state = %ld bytes:\n", sizeof(sponge.state));
    for (int i=0; i<sizeof(csi.sponge.state); i++) printf ("%2.2x ", csi.sponge.state[i]); printf ("\n");
    printf("      sponge.rate = %d bits, %d bytes\n", csi.sponge.rate, csi.sponge.rate/8);
    printf("      sponge.byteIOIndex = %d\n", csi.sponge.byteIOIndex);
    printf("      sponge.squeezing = %d\n", csi.sponge.squeezing);
    printf ("  csi.fixedOutputLength = %ld\n", csi.fixedOutputLength);
    printf ("  csi.lastByteBitLen = %d\n", csi.lastByteBitLen);
    printf ("  csi.lastByteValue = %d\n", csi.lastByteValue);
    printf ("  csi.emptyNameCustom = %d\n", csi.emptyNameCustom);
    printf ("  csi.phase = %d\n", csi.phase);
    printf("KMAC additional parameters:\n");
    printf("km->outputBitLen = %ld\n", km->outputBitLen);

    return;
}

void print_EXTRA (EXTRA_DATA extra)
{
    printf ("EXTRA_DATA\n");

    printf("magic = %llx\n", extra->magic);
    printf("base64flag = %d\n", extra->base64flag);
    printf("xofflag = %d\n", extra->xofflag);
    printf("xoflength = %lld\n", extra->xoflength);
    printf("this_chunk_size, more_size = %ld, %lld\n", extra->this_chunk_size,
            extra->more_size);
    printf("KMAC additional parameters:\n");
    printf ("key \"%s\", length %ld bytes\n", extra->key, extra->KeyBitLen>>3);
    printf ("customize string \"%s\", length %ld bytes\n", extra->customize, extra->customizeBitLen>>3);
    printf("outputBitLen = %ld\n", extra->outputBitLen);

    return;
}


/**************************** KMACXOF128 ***************************/
HashReturn myKMACXOF128_init (hashState  *state, int hashbitlen, EXTRA_DATA extra)
{
    /* verify correct hash length   */
    if (hashbitlen != HASH_BITLENGTH_KMAC_128)
        return BAD_HASHBITLEN;

    const       BitSequence *key    = extra->key;
    BitLength   keyBitLen           = extra->KeyBitLen;
    BitLength   outputBitLen        = HASH_BITLENGTH_KMAC_128;
    const BitSequence *customization = extra->customize;
    BitLength   customizeBitLen     = extra->customizeBitLen;
    
    /* allocate context and fill it */
    SHA3_CTX *context   = (SHA3_CTX *)malloc (sizeof (SHA3_CTX));
    memset (context, 0, sizeof (SHA3_CTX));
    context->hashbitlen = HASH_BITLENGTH_KMAC_128;      // 256 bits, 32 bytes
    context->magic      = HASH_MAGIC_KMAC_128;
    context->xof_OK     = 1;
    context->xoflength  = HASH_BITLENGTH_KMAC_128;      // 256 bits, 32 bytes
    context->xofinfinity = 0;
    context->xofflag    = 0;
    // 
    context->this_chunk_size = HASH_LENGTH_KMAC_128;  // 32 bytes, 256 bits
    context->more_size = HASH_LENGTH_KMAC_128;          // 32 bytes, 256 bits

    /* extra data for XOF */
    if (extra) {
        context->base64flag = extra->base64flag;
        context->xofflag    = extra->xofflag;
        context->xoflength  = extra->xoflength;
        context->this_chunk_size = HASH_LENGTH_KMAC_128;
        context->more_size  = extra->xoflength;
        context->xofinfinity = (context->xoflength == -1);
        // default length is  32 bytes if xoflength == 0 (unspecified)
        if (extra->xoflength == 0) {    // use default length
            context->xoflength = HASH_LENGTH_KMAC_128;    // 32 bytes
            context->more_size = HASH_LENGTH_KMAC_128;   //  32 bytes
            context->this_chunk_size = HASH_LENGTH_KMAC_128;   //  32 bytes
        }
        // length is shorter than default: do not squeeze
        if ((context->xoflength > 0) && (context->xoflength < HASH_LENGTH_KMAC_128)) { 
            context->more_size = context->this_chunk_size = context->xoflength;
        }
        // length is infinity
        if (extra->xoflength == -1) {     // -1, infinity
            context->xoflength= -1;
            context->more_size = HASH_LENGTH_KMAC_128;  // 32 bytes, 256 bits
            context->xofinfinity = 1;
        }
        context->binoutflag = extra->binoutflag;    // not yet used
    }

    KMAC_Instance *km   = &context->km;
    
	*state = (hashState *) context;

/* Ronny Van Keer <rvankeer@yahoo.com> writes on 18.11.2024:

Dear Harald,

If you call init with a non zero output length, you can't call squeeze 
after final.
So, if you want to call squeeze multiple times, you have to 
call init with zero length, and then you should able to 
squeeze multiple times after the final call.

Best regards,
Ronny
 */

    int retval =  KMACXOF128_Initialize(km, key, keyBitLen, 0L, 
                    customization, customizeBitLen);
    return retval;

}

HashReturn  myKMACXOF128_update (
    hashState state,          /* previously initialized context */
    const BitSequence *buffer,  /* bit buffer, first bit is MSB in [0] */
    DataLength databitlen)      /* number of bits to process from buffer */
{
    /* can be called once or many times */
    /* verify correct hashbitlen and magic  */

    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_KMAC_128)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_KMAC_128)
        return BAD_ALGORITHM;

    KMAC_Instance *km   = &context->km;

	int retval = KMACXOF128_Update (km, buffer, databitlen);
    return retval;
}

HashReturn  myKMACXOF128_final (hashState state, BitSequence *hashval)
{
    SHA3_CTX *context = (SHA3_CTX *) state;

    HashReturn  retval;

    if (context->hashbitlen != HASH_BITLENGTH_KMAC_128)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_KMAC_128)
        return BAD_ALGORITHM;

    KMAC_Instance *km   = &context->km;

    /* check if already squeezing   */

    if (!context->squeezeflag) {    // first call produces no output
        retval = KMACXOF128_Final (km, context->out);
    }
    // squeezing
    context->squeezeflag++;

     /* total length in bytes is in context->xoflength, 
      * remaining length in context->more_size
      *  BitLength   outputBitLen        = HASH_BITLENGTH_KMAC_128;
      * -1 means infinity    */
    BitLength obl = HASH_BITLENGTH_KMAC_128;
    if (context->more_size<HASH_BITLENGTH_KMAC_128) obl = context->more_size*8;
    retval = KMACXOF128_Squeeze (km, context->out, obl);
    if (hashval) memmove (hashval, context->out, obl/8);

    if (!context->xofinfinity) {
        // not last block of data
        if (context->more_size > context->this_chunk_size) {
            context->more_size -= context->this_chunk_size;
        } else {    // last block of data
            context->this_chunk_size = context->more_size;
            context->more_size = 0;
        }
    }

    // infinity means: more_size remains on -1, 
    // this_chunk_size remains on its standard value 

    return retval;
}

/* Digests a file and prints the result.
 */
HashReturn myKMACXOF128_File (hashState state, FILE *in)
{
	SHA3_CTX *context = (SHA3_CTX *) state;
	int len;
	unsigned char buffer[BUFFERSIZE];
	HashReturn retval;
    char    printbuf[BUFFERSIZE];
    memset(printbuf, 0, BUFFERSIZE); // initialize to find memory leaks


	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_KMAC_128)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_KMAC_128)
		return BAD_ALGORITHM;

    KMAC_Instance *km   = &context->km;

	while ((len = fread (buffer, 1, BUFFERSIZE, in))) {
		retval = myKMACXOF128_update (context, buffer, (DataLength)len << 3);
		if (retval != SUCCESS) return retval;
	}

    int offset = 0;
	
	retval = myKMACXOF128_final (context, context->out);
    // copy first buffer into printbuf
    memmove (printbuf, context->out, context->this_chunk_size);
    offset += context->this_chunk_size;
#if 0
    for (int i = 0; i<offset; i++) printf ("%2.2x", context->out[i]&0xff);     
    printf ("\n");
    for (int i = 0; i<offset; i++) printf ("%2.2x", printbuf[i]&0xff);     
    printf ("\n");
#endif
    while (context->more_size && context->xofflag) {
        retval = myKMACXOF128_final (context, context->out);
        // copy next buffer into printbuf
        memmove (printbuf+offset, context->out, context->this_chunk_size);
        offset += context->this_chunk_size;
    }
#if 0
    for (int i = 0; i<offset; i++) printf ("%2.2x", printbuf[i]&0xff);     
    printf ("\n");
#endif
	fclose (in);
    // copy printbuf back, but context->out is only 600 bytes long!
    int bytescopied = offset<sizeof(context->out)? offset : sizeof(context->out);
    memmove (context->out, printbuf, bytescopied);

	return retval;
}


HashReturn myKMACXOF128_HashToByte (hashState state, BYTE *out) 
{
	SHA3_CTX *context = (SHA3_CTX *) state;

	/* verify correct hashbitlen and magic	*/
    if (context->hashbitlen != HASH_BITLENGTH_KMAC_128)
        return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_KMAC_128)
		return BAD_ALGORITHM;

    int len = HASH_LENGTH_KMAC_128;
    if (context->xoflength > 0) len = context->xoflength;
	memcpy (out, context->out, len);

	return SUCCESS;
}

#if 0
HashReturn myKMACXOF128_hash (int hashbitlen, const BitSequence *data,
                      DataLength databitlen, BitSequence *hashval)
{
    /* keystring, keylength, customstring, customBitLen
     * are global variables, I found no other way to feed them into
     * this function - hvf 23.09.2024
     */
    HashReturn  retval;
    KMAC_Instance   km;
    retval =  KMACXOF128_Initialize(&km, keystring, keylength, 
                HASH_BITLENGTH_KMAC_128, customstring, customBitLen);
    
    if (retval != SUCCESS) {
        fprintf (stderr, "myKMACXOF128_init failed, reason %d, hash length %d\n",
                 retval, HASH_BITLENGTH_KMAC_128);
        exit (1);
    }

    retval = KMACXOF128_Update (&km, data, databitlen);
    if (retval != SUCCESS) {
        fprintf (stderr, "myKMAC128_update failed, reason %d\n", retval);
        exit (1);
    }

    /* result is returned in hashval    */
    retval = KMACXOF128_Final (&km, hashval);
    if (retval != SUCCESS) {
        fprintf (stderr, "myKMAC128_final failed, reason %d\n", retval);
        exit (1);
    }
    return retval;
}
#endif



/**************************** KMACXOF256 ***************************/
HashReturn myKMACXOF256_init (hashState  *state, int hashbitlen, EXTRA_DATA extra)
{
    /* verify correct hash length   */
    if (hashbitlen != HASH_BITLENGTH_KMAC_256)
        return BAD_HASHBITLEN;

    // KMAC_Instance *km = &(SHA3_CTX *state)->km;
    const       BitSequence *key    = extra->key;
    BitLength   keyBitLen           = extra->KeyBitLen;
    BitLength   outputBitLen        = HASH_BITLENGTH_KMAC_256;
    const BitSequence *customization = extra->customize;
    BitLength   customizeBitLen     = extra->customizeBitLen;
    
    /* allocate context and fill it */
    SHA3_CTX *context   = (SHA3_CTX *)malloc (sizeof (SHA3_CTX));
    memset (context, 0, sizeof (SHA3_CTX));
    context->hashbitlen = HASH_BITLENGTH_KMAC_256;
    context->magic      = HASH_MAGIC_KMAC_256;
    context->xof_OK     = 1;
    context->xoflength  = HASH_BITLENGTH_KMAC_256;      // 512 bits, 64 bytes
    context->xofinfinity = 0;
    context->xofflag    = 0;
    
    context->this_chunk_size = HASH_LENGTH_KMAC_256;    // 64 bytes, 512 bits
    context->more_size = HASH_LENGTH_KMAC_256;          // 64 bytes, 512 bits

    /* extra data for XOF */
    if (extra) {
        context->base64flag = extra->base64flag;
        context->xofflag    = extra->xofflag;
        context->xoflength  = extra->xoflength;
        context->this_chunk_size = HASH_LENGTH_KMAC_256;
        context->more_size  = extra->xoflength;
        context->xofinfinity = (context->xoflength == -1);
        // default length is  64 bytes if xoflength == 0 (unspecified)
        if (extra->xoflength == 0) {    // use default length
            context->xoflength = HASH_LENGTH_KMAC_256;    // 64 bytes, 512 bits
            context->more_size = HASH_LENGTH_KMAC_256;    // 64 bytes, 512 bits
            context->this_chunk_size = HASH_LENGTH_KMAC_256;   // 64 bytes
        }
        // length is shorter than default: do not squeeze
        if ((context->xoflength > 0) && (context->xoflength < HASH_LENGTH_KMAC_256)) {
            context->more_size = context->this_chunk_size = context->xoflength;
        }
        // length is infinity
        if (extra->xoflength == -1) {     // -1, infinity
            context->xoflength= -1;
            context->more_size = HASH_LENGTH_KMAC_256;  // 64 bytes, 512 bits
            context->xofinfinity = 1;
        }

        context->binoutflag = extra->binoutflag;    // not yet used
    }

    KMAC_Instance *km   = &context->km;
    
	*state = (hashState *) context;

    int retval =  KMACXOF256_Initialize(km, key, keyBitLen, 0L, 
                    customization, customizeBitLen);
    return retval;

}



HashReturn  myKMACXOF256_update (
    hashState state,          /* previously initialized context */
    const BitSequence *buffer,  /* bit buffer, first bit is MSB in [0] */
    DataLength databitlen)      /* number of bits to process from buffer */
{
    /* can be called once or many times */
    /* verify correct hashbitlen and magic  */

    SHA3_CTX *context = (SHA3_CTX *) state;
    if (context->hashbitlen != HASH_BITLENGTH_KMAC_256)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_KMAC_256)
        return BAD_ALGORITHM;

    KMAC_Instance *km   = &context->km;

	int retval = KMACXOF256_Update (km, buffer, databitlen);
    return retval;
}

HashReturn  myKMACXOF256_final (hashState state, BitSequence *hashval)
{
    SHA3_CTX *context = (SHA3_CTX *) state;

    HashReturn  retval;

    if (context->hashbitlen != HASH_BITLENGTH_KMAC_256)
        return BAD_HASHBITLEN;

    if (context->magic != HASH_MAGIC_KMAC_256)
        return BAD_ALGORITHM;

    KMAC_Instance *km   = &context->km;

    /* check if already squeezing   */

    if (!context->squeezeflag) {    // skip if squeezing
        retval = KMACXOF256_Final (km, context->out);
    }
    // squeezing
    context->squeezeflag++;
    BitLength obl = HASH_BITLENGTH_KMAC_256;
    if (context->more_size<HASH_BITLENGTH_KMAC_256) obl = context->more_size*8;
    retval = KMACXOF256_Squeeze (km, context->out, obl);
    if (hashval) memmove (hashval, context->out, obl/8);

    if (!context->xofinfinity) {
        // not last block of data
        if (context->more_size > context->this_chunk_size) {
            context->more_size -= context->this_chunk_size;
        } else {    // last block of data
            context->this_chunk_size = context->more_size;
            context->more_size = 0;
        }
    }
    // infinity means: more_size remains on -1, 
    // this_chunk_size remains on its standard value
    return retval;
}

/* Digests a file and prints the result.
 */
HashReturn myKMACXOF256_File (hashState state, FILE *in)
{
	SHA3_CTX *context = (SHA3_CTX *) state;
	int len;
	unsigned char buffer[BUFFERSIZE];
	HashReturn retval;
    char    printbuf[BUFFERSIZE];
    memset(printbuf, 0, BUFFERSIZE); // initialize to find memory leaks


	/* verify correct hashbitlen and magic	*/
	if (context->hashbitlen != HASH_BITLENGTH_KMAC_256)
		return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_KMAC_256)
		return BAD_ALGORITHM;

    KMAC_Instance *km   = &context->km;

	while ((len = fread (buffer, 1, BUFFERSIZE, in))) {
		retval = KMACXOF256_Update (km, buffer, (DataLength)len << 3);
		if (retval != SUCCESS) return retval;
	}

    int offset = 0;

	retval = myKMACXOF256_final (context, context->out);
    // copy first buffer into printbuf
    memmove (printbuf, context->out, context->this_chunk_size);
    offset += context->this_chunk_size;
#if 0
    for (int i = 0; i<offset; i++) printf ("%2.2x", context->out[i]&0xff);     
    printf ("\n");
    for (int i = 0; i<offset; i++) printf ("%2.2x", printbuf[i]&0xff);     
    printf ("\n");
#endif
    while (context->more_size && context->xofflag) {
        retval = myKMACXOF256_final (context, context->out);
        // copy next buffer into printbuf
        memmove (printbuf+offset, context->out, context->this_chunk_size);
        offset += context->this_chunk_size;
    }
#if 0
    for (int i = 0; i<offset; i++) printf ("%2.2x", printbuf[i]&0xff);     
    printf ("\n");
#endif
    fclose (in);
    // copy printbuf back, but context->out is only 600 bytes long!
    int bytescopied = offset<sizeof(context->out)? offset : sizeof(context->out);
    memmove (context->out, printbuf, bytescopied);
    
	return retval;
}


HashReturn myKMACXOF256_HashToByte (hashState state, BYTE *out) 
{
	SHA3_CTX *context = (SHA3_CTX *) state;

	/* verify correct hashbitlen and magic	*/
    if (context->hashbitlen != HASH_BITLENGTH_KMAC_256)
        return BAD_HASHBITLEN;

	if (context->magic != HASH_MAGIC_KMAC_256)
		return BAD_ALGORITHM;

    int len = HASH_LENGTH_KMAC_256;
    if (context->xoflength > 0) len = context->xoflength;
    memcpy (out, context->out, len);
	
	return SUCCESS;
}

#if 0
HashReturn myKMACXOF256_hash (int hashbitlen, const BitSequence *data,
                      DataLength databitlen, BitSequence *hashval)
{
    HashReturn  retval;
    KMAC_Instance   km;
    retval =  KMACXOF256_Initialize(&km, keystring, keylength, 
                HASH_BITLENGTH_KMAC_256, customstring, customBitLen);
    
    if (retval != SUCCESS) {
        fprintf (stderr, "myKMACXOF256_init failed, reason %d, hash length %d\n",
                 retval, HASH_BITLENGTH_KMAC_256);
        exit (1);
    }

    retval = KMACXOF256_Update (&km, data, databitlen);
    if (retval != SUCCESS) {
        fprintf (stderr, "myKMAC256_update failed, reason %d\n", retval);
        exit (1);
    }

    /* result is returned in hashval    */
    retval = KMACXOF256_Final (&km, hashval);
    if (retval != SUCCESS) {
        fprintf (stderr, "myKMAC256_final failed, reason %d\n", retval);
        exit (1);
    }
    return retval;
}
#endif

