/*
 * Generic Hash and HMAC Program
 *
 * Copyright (C) 2009 Harald von Fellenberg <hvf@hvf.ch>
 *
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 3 of the License, or 
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */

/* 26.02.2015 hvf	allow base64 encoded hashes, e.g. shake128-b64 (file) = ...
 * 01.05.2015 hvf	allow variable length for XOFSHAKEnnn functions
 * 13.05.2015 hvf	KMAC implemented
 *					show algorithm on each result (OK, differ)
 */
/* 31.08.2024 hvf   HMAC also for sha3-nnn, shakennn algorithms 
 *                  see NIST SP 800-224, June 2024
 *                  added --hmacverbose flag for HMAC, KMAC output
 *  29.09.2024 hvf  added kmac128, kmac256, kmacxof128, kmacxof256
 */

#include	"fehashmac.h"
#include	<stdio.h>
#include	"fehashmac-macros.h"
#include	<assert.h>
#include	"base64.h"

// kmac128, kmac256 parameters
extern  int     customflag;
extern  char    *customstring;

char *strsave (char *in); /* save a string and return its pointer */

char *stripquotes (char *in) /* take away leading and trailing quotes */
						/* the pointer may be freed after use */
{
	char *buf = NULL;
	if (!in) return NULL;
	size_t len = strlen (in);
	buf = malloc (len+1);
	if (!buf) {
		perror ("strsave");
		exit (1);
	}
	memset (buf, 0, len+1);
    int foundquote = 0;
    if (in[0] == '"') {             // first quote found: step over it
	    strncpy (buf, in+1, len);
        foundquote = 1;
    }
	else strncpy (buf, in, len+1);
    if (foundquote) {
        for (int i = 0; i<strlen(buf); i++) {
            if (buf[i] == '"') {    // second quote ends the string
                buf[i] = '\0';
                break;
            }
        }
    } else {
        for (int i = 0; i<strlen(buf); i++) {
            if (isspace (buf[i])) {
                buf[i] = '\0';
                break;
            }
        }
    }
	return buf;
}

/* NOTE: HASHBUFSIZ is at least 4k and replaces BUFSIZ, 
 * which is only 1k on certain OS's like FreeBSD, Solaris
 */

// read a file with digests and verify them

void ReadDigest (char *listfilename)
{
	FILE 	*listfile;
	char	buf[HASHBUFSIZ];
	char	printbuf[HASHBUFSIZ];
	char	tempbuf[HASHBUFSIZ];
    /* the 3 elements of each line: keyword (object) = value */
	char	*keyword;
	char	*object;
	char	*value;
	char	*algoptr;
	char	*filename;
	FILE	*in;
	char	*filedigest;
	char	*ptr;
	BYTE	out[HASHBUFSIZ];
	int		i, j;
	int		length;
    int     retval;
	struct	hash_algo	*a;
	void	*context;
    // parsing indicators
    int     foundkeyword;
	int		foundalgo, foundleftp, foundrightp, foundfilename, founddigest;
    int     foundquote;
	char	*hmacprefix = NULL;
	char	*kmacprefix = NULL;
	int		do_hmac = 0;	// HMAC for current file requested
	int		do_kmac = 0;	// KMAC for current file requested
	struct	use_algo use;
	int		ok = 0, failed = 0, total = 0, found = 0, notfound = 0, hmacfound = 0;
	int		kmacfound = 0;
	int		kmackeyfound = 0;
	int		kmacxxxfound = 0;   // kmac128, kmacxof256 etc.
	int		customfound = 0;    // for kmac128 etc.

	int		lines = 0, algofound = 0, algonotfound = 0;
	long long xoflength = 0;	// length (xof) = nnn for xofshakexxx
	int		xofflag = 0;
	int		xoffound = 0;
	int		keylengthfound = 0;
	char	*displayname = NULL;

    char    *mystring = NULL; 
    char    *mynewstring = NULL;
    char    *myvalue = NULL;  
    char    *mynewvalue = NULL;

    char    *savekeystring = NULL;
    char    *savehexkeystring = NULL;
    char    *savecustomstring = NULL;
	
    struct  extra extra;
    memset (&extra, 0, sizeof(extra));

	hmacflag = kmacflag = keyflag = hexkeyflag = 0;

	if (!listfilename) listfile = stdin;
	else {
		listfile = fopen (listfilename, "rb");
		if (!listfile) {
			perror (listfilename);
			exit (1);
		}
	}
	/* each line has the format:
     * keyword (object) = value
     * keywords:
     * key, customstring, simplealgo, hmac-simplealgo, kmac-simplealgo,
     * simplealgo-b64
     * object: 
     * filename e.g. /dev/null
     * string, "string"
     * value: 
     * digest, valuestring, "value string"
	 * algorithm (filename) = digest
	 * algorithm may be: simple, hmac-simple, simple-b64
	 * we look for algoptr, filename, filedigest
	 * i.e. a line looks like
	 * algoptr (filename) = filedigest
     * or:  keyword (object) = "value"
     * or:  keyword ("string") = value
	 */

	while (fgets (buf, sizeof(buf), listfile)) {
//printf ("line: %s", buf);
        // skip lines starting with #
        if (buf[0] == '#') continue;
		lines++;
		memset (&use, 0, sizeof (use));
		memset (printbuf, 0, sizeof (printbuf));
		memset (tempbuf, 0, sizeof (printbuf));
		memset (out, 0, sizeof (out));
		// delete trailing newline
		if (buf[strlen(buf)-1] == '\n') buf[strlen(buf)-1] = '\0';
        char *savebuf = strsave(buf);
		foundalgo=foundleftp=foundrightp=foundfilename=founddigest=0;
        foundquote = 0;
        foundkeyword =0;
		if ((length = strlen (buf)) == (sizeof(buf) -1)) {
			fprintf (stderr, "input too long, exiting.\n");
			exit (1);
		}
		filename = NULL;
		filedigest = NULL;
		algoptr = buf;		// no leading white space allowed
		keyword = buf;		// no leading white space allowed
		for (ptr=buf, i=0; i<length; i++, ptr++) {
			if (isspace (*ptr)) {	// find terminating white space or '('
				*ptr++ = '\0';
				break;
			} else if (*ptr == '('){	// no blank after algo
				*ptr++ = '\0';
				// filename = ptr++;
                // mystring = filename;
				object = ptr++;
                filename = object;
                mystring = object;

				foundleftp=1;
				break;
			}
		}
		//foundalgo = strlen (algoptr);
		foundkeyword = strlen (keyword);
		if (!keyword) continue;		// try next line
		if (!foundleftp) {				// look for opening parenthesis
			for (i=ptr-buf; i<length; i++, ptr++) {
				if (*ptr == '('){
					//filename = ++ptr;	// filename starts after opening par
					object = ++ptr;	// filename starts after opening par
                    filename = object;
					foundleftp=1;
					break;
				}
			}
		}
		if (!foundleftp) continue;		// try next line
		for (i=ptr-buf; i<length; i++, ptr++) {
			if (*ptr == ')'){		// right parenthesis ends filename
				*ptr++ = '\0';
				foundrightp=1;
                    //mystring = filename;
                    mystring = object;
                    mynewstring = stripquotes (mystring); // elim quotes
				break;
			}
		}
//printf ("object = %s\n", object);
		if (!foundrightp) continue;
		for (i=ptr-buf; i<length; i++, ptr++) {
			if (!isspace (*ptr) && *ptr != '='){	// skip ' = '
				//filedigest = ptr;	// filedigest starts here
				value = ptr;	// filedigest starts here
                filedigest = value;
                myvalue = filedigest;
                mynewvalue = stripquotes (myvalue); // elim quotes
				break;
			}
		}
//printf ("value = %s\n", value);
		// find end of value
		for (i=ptr-buf; i<length; i++, ptr++) {
            if (isspace(*ptr)) {	// space, newline end the digest
				*ptr = '\0';
				break;
			}
		}
//printf ("value = %s\n", value);

//printf ("mynewstring = %s, mynewvalue = %s\n", mynewstring, mynewvalue);
		// if (!algoptr || !filename || !filedigest) 
		// if (!keyword || !mynewstring || !mynewvalue) 
		if (!keyword || !object || !value) {
			fprintf (stderr, "keyword %s, object %s, value %s Incomplete, input line, skipping.\n", keyword, object, value);
			continue;
		}

		/* lowercase algorithm, just in case */
		//for (ptr=algoptr; *ptr; ptr++)
		for (ptr=keyword; *ptr; ptr++)
			*ptr = tolower (*ptr);

		// correct line found, count it
		total++;

        // printf ("input line = \'%s\'\n", savebuf);
        // printf ("key %s, object %s, value \'%s\'\n", algoptr, mynewstring, mynewvalue);

		// special record for shakexxx: length (xof) = nnn
        // and hmac-length (xof) = nnn

		if (!strcmp (keyword, "xoflength")) {
			if (!strcmp (object, "Bytes")) {
				// length (xof) = found
				xofflag++;
				xoffound++;
				xoflength = atoll (mynewvalue);
            }
            continue;
        }

		if (!strcmp (keyword, "length") | !strcmp (keyword, "hmac-length")) {
			// length key found
			//if (!strcmp (filename, "xof")) 
			if (!strcmp (object, "xof")) {
				// length (xof) = found
				xofflag++;
				xoffound++;
				xoflength = atoll (mynewvalue);
			} else if (!strcmp (mynewstring, "kmac")) { // length (kmac) = nnn
				keylength = atoi (mynewvalue);
				keylengthfound++;
			} else {
				fprintf (stderr, "value \"%s\" for length not known.\n", mynewstring);
				algonotfound++;
			}
			continue;
		}

		// special record for HMAC, KMAC: key (hmac) = ... -and- hexkey (hmac) = ...
        if (!strcmp (keyword, "customstring")) {
            customstring = mynewvalue;
            customflag = 1;
            continue;
        }

        // hexkey (hmac) = nnnn
        // Key (hmac) = abcd
        // kmac128 (file) = digest
        // kmac256 ("string") = digest

		if (!strcmp (keyword, "key") || !strcmp (keyword, "hexkey")) {
			if (!strcmp (object, "hmac")) {
				// hmac key found, count it
				hmacfound++;
				hmacprefix = strdup (mynewstring);
				hmacflag = 1;
				hmacprintflag = 0;
			}
			if (!strcmp (object, "kmac")) {
				// kmac key found, count it
				kmacfound++;
				kmacprefix = strdup (mynewstring);
				kmacflag = 1;
				kmacprintflag = 0;
			}
			keystring = hexkeystring = strdup (mynewvalue);
            extra.key = keystring;  // OK for ASCII key, see below
			keyflag = !strcmp (algoptr, "key");
			//if (keyflag) keylength = 0;	// ASCII keys for KMAC do not support keylength
            keylength = strlen(keystring);
            extra.KeyBitLen = keylength<<3;
			hexkeyflag = !strcmp (keyword, "hexkey");
            if (hexkeyflag) {   // convert keystring into hex
                savehexkeystring = strdup (keystring);
                int nibbles = copyhextobin (savehexkeystring, keylength, 
                                    keystring, keylength);
                if (nibbles & 0x01) {
                    fprintf (stderr, "do_HexString: read odd number of hex values: %d, last bits may have wrong alignment\n", nibbles);
                }
                extra.key = savehexkeystring;
                extra.KeyBitLen = nibbles*4;
            }
			continue;
		}

		/* keyword might be the alorithm incl pre- and postfix, save it	*/
		displayname = strsave (keyword);

        if (!strncmp (keyword, "hmac-", strlen ("hmac-"))) {
            algoptr += strlen ("hmac-");
            do_hmac = 1;
            do_kmac = 0;
        }
        if (!strncmp (keyword, "kmac-", strlen ("kmac-"))) {
            algoptr += strlen ("kmac-");
            do_kmac = 1;
            do_hmac = 0;
        }
//printf ("readdigest: keyword, object, value = %s %s %s\n", keyword, object, value);
		// remove base64 suffix (-b64)
		int	myb64 = 0;
        char *p; 
        if ((p = strstr (algoptr, "-b64"))) { // trailing -b64
            myb64 = 1; 
            *p = '\0'; /* delete trailing -b64 */ 
        } 

		// now look for the right algorithm
		// printf ("algoptr = %s\n", algoptr);
		for (a=NULL, j=0; j<HashTableSize; j++) {
			if (!strcmp (algoptr, HashTable[j].name)){
				a = HashTable + j;
				algofound++;
				break;
			}
		}
		if (!a) {
			fprintf (stderr, "Algorithm \"%s\" not found, skipping.\n", algoptr);
			algonotfound++;
			continue;
		}

        /* check that xoflength is not too big, maximum is 600 bytes */
        if (xoflength > SqueezingOutputLength/8) {
            printf ("xoflength %lld too big, processing stops.\n", xoflength);
            exit (1);
        }
		use.a = a;
		extra.xofflag = xofflag;
		use.extra.xofflag = xofflag;
        extra.xoflength = xoflength;
        use.extra.xoflength = xoflength;
        extra.this_chunk_size = xoflength<32? xoflength : 32;
        use.extra.this_chunk_size = extra.this_chunk_size;
        extra.more_size = xoflength;
        use.extra.more_size = xoflength;

    	extra.customize = customstring; 
		extra.customizeBitLen = strlen(customstring) << 3; 
#if 0
printf ("algoptr = %s\n", algoptr);
printf ("extra.key = %s\n", extra.key);
printf ("extra.KeyBitLen = %ld\n", extra.KeyBitLen);
printf ("extra.customize, customizeBitLen = %s, %ld\n", extra.customize, extra.customizeBitLen);
//print_EXTRA(&extra);
#endif	
		// we decode the base64 filedigest, so no extra encoding
		// use.extra.base64flag = myb64;
		extra.base64flag = 0;
		retval = (*a->init)(&context, a->hash_length<<3, &extra);
//printf ("%s init: retval = %d\n", a->name, retval);
		if (!(in = fopen (filename, "rb"))) {
			notfound++;
			perror (filename);
			continue;
		}
		found++;
		if (do_hmac) {
// printf ("doing hmac...\n");
			HMAC_INIT ((&use), keyflag, hmacprintflag);
			HMAC_UPDATE ((&use), context);
			hmacprintflag = 0;	// once is enough
		}
		if (do_kmac) {
// printf ("doing kmac...\n");
			KMAC_INIT ((&use), keyflag, kmacprintflag);
			KMAC_UPDATE ((&use), context);
			kmacprintflag = 0;	// once is enough
		}
// #define kmacxofxxx_OK(bitmask) ((bitmask) & 0x08)
        if (kmacxofxxx_OK(a->mac_OK))printf ("xof flag seen\n");

		// digest file - init has already been called
		retval = (*a->file)(context, in);
        int len = a->hash_length;
        if (kmacxofxxx_OK(a->mac_OK)) {
            SHA3_CTX *ctx3 = (SHA3_CTX *)context;
            if (ctx3->xoflength > 0) len = ctx3->xoflength;
            len = xoflength;
        }
		// make sure out is large enough
		assert (sizeof(out) >= len);
		(*a->hashtobyte)(context, out);
        // doing hmac final
		if (do_hmac) {
			HMAC_FINAL ((&use), out);
		}

		// decompress base64 hash if required
		if (myb64) {
			// check if supported
			if (!a->base64_OK) {
				fprintf (stderr, "%s-b64 notsupported, check skipped.\n", a->name);
				continue;
			}

			// decompress base64 filedigest
			/* bool ok = base64_decode_alloc (in, inlen, &out, &outlen);
			 * if (!ok)
			 *   FAIL: input was not valid base64
			 * if (out == NULL)
			 *   FAIL: memory allocation error
			 * OK: data in OUT/OUTLEN
			 */
			char *unb64out = NULL;
			size_t unb64outlen = 0;
			bool ok = base64_decode_alloc (filedigest, strlen (filedigest), 
                    &unb64out, &unb64outlen);
			if (!ok) {
				fprintf (stderr, "readdigest: input is not a valid base64 string: %s\n",
						filedigest);
				exit (1);
			}
			if (unb64out == NULL) {
                fprintf (stderr, "base64  memory allocation error.\n");
                exit (1);
            }
			// ok, data is in unb64out
			for (i=0, j=0; i<unb64outlen && j<HASHBUFSIZ-3; i++) {
				j += snprintf (tempbuf+j, 3, "%2.2x", unb64out[i]&0xff);
			}
			// filedigest is now in tempbuf
			filedigest = tempbuf;
			if (unb64out) free (unb64out); unb64out = NULL;
		} 
			// normal hashes pass here
			if (!a->xof_OK_defaultLength) {
				for (i=0, j=0; i<len && j<HASHBUFSIZ-3; i++) {
					j += snprintf (printbuf+j, 3, "%2.2x", out[i]&0xff);
				}
			} else {
				//for (i=0, j=0; i<a->hash_length && j<a->xof_OK_defaultLength*2 && j<HASHBUFSIZ-3; i++) 
				for (i=0, j=0; i<len && j<HASHBUFSIZ-3; i++) {
					j += snprintf (printbuf+j, 3, "%2.2x", out[i]&0xff);
				}
			}
#undef DEBUG
#ifdef DEBUG
		printf ("%s calc: %s\n", displayname, printbuf);
		printf ("%s read: %s\n\n", displayname, filedigest);
		printf ("%s calc hash length: %ld\n", displayname, strlen (printbuf)/2);
		printf ("%s read hash length: %ld\n", displayname, strlen (filedigest)/2);
#endif
		size_t lenfile = strlen (filedigest);
		size_t lenprint = strlen (printbuf);
		if (strcmp (filedigest, printbuf) == 0) {
			
			printf ("%s: OK %s\n", filename, displayname);

			ok++;
		}
		
		/* interesting: the shortened string comparison applies to the 
		 * first parameter only, so we need to do two separate comparisons
		 * to cover both cases: digest shorter than calculated value, and
		 * calculated value shorter than digest
		 */
		else if (strncmp (filedigest, printbuf, lenfile) == 0) {	// digest is shorter
			printf ("%s: OK %s, match truncated to %ld bytes\n", filename, displayname, lenfile/2);
			ok++;
		}
		else if (strncmp (printbuf, filedigest, lenprint) == 0) {	// calculated is shorter
			printf ("%s: OK %s, match truncated to %ld bytes\n", filename, displayname, lenprint/2);
			ok++;
		}
		else { 
			printf ("%s: differ %s\n", filename, displayname);
			failed++;
		}
		(*a->free)(context); context = NULL;
		if (displayname) { free (displayname); displayname = NULL; }
	}
	fclose (listfile);
	if (hmacprefix) { hmacprefix = NULL; }
	if (kmacprefix) { kmacprefix = NULL; }
	if (keystring)  { keystring = hexkeystring = NULL; }
	/*
	 * we print only values different from zero to shorten output
	 * 17.05.2015
	 */
	printf ("\nSummary:\n");
	printf ("--------\n");
	if (lines>0) 			printf ("Lines read:         %5d\n", lines);
	if (total>0) 			printf ("Hash entries read:  %5d\n", total);
	if (hmacfound>0) 		printf ("HMAC keys found:    %5d\n", hmacfound);
	if (kmacfound>0) 		printf ("KMAC keys found:    %5d\n", kmacfound);
	if (keylengthfound>0) 	printf ("KMAC key lengths:   %5d\n", keylengthfound);
	if (xoffound>0) 		printf ("XOF lengths found:  %5d\n", xoffound);
	if (found>0) 			printf ("Files found:        %5d\n", found);
	if (notfound>0) 		printf ("Files not found:    %5d\n", notfound);
	// printf ("Algorithms found: %5d\n", algofound);
	if (algonotfound>0) 	printf ("Unknown Algorithms: %5d\n", algonotfound);
	if (ok>0) 				printf ("Files OK:           %5d\n", ok);
	if (failed>0) 			printf ("Files failed:       %5d\n", failed);
}

char *strsave (char *in) /* save a string and return its pointer */
						/* the pointer may be freed after use */
{
	char *buf = NULL;
	if (!in) return NULL;
	size_t len = strlen (in);
	buf = malloc (len+1);
	if (!buf) {
		perror ("strsave");
		exit (1);
	}
	memset (buf, 0, len+1);
	strncpy (buf, in, len+1);
	return buf;
}

