/*
 * Generic Hash and HMAC Program
 *
 * Copyright (C) 2009 Harald von Fellenberg <hvf@hvf.ch>
 *
 * This program is free software; you can redistribute it and/or modify 
 * it under the terms of the GNU General Public License as published by 
 * the Free Software Foundation; either version 3 of the License, or 
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful, but 
 * WITHOUT ANY WARRANTY; without even the implied warranty of 
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. 
 * See the GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License 
 * along with this program; if not, see <http://www.gnu.org/licenses/>.
 */
/*  added kmac test program - hvf 23.10.2024    
 */

#include	"fehashmac.h"
#include	"fehashmac-macros.h"
#include	<stdio.h>
#include	<assert.h>

#if 0
struct hash_algo {
    /* the name of the hash algorithm */
    char *name;
    /* the init function, inits a ptr   */
    /* extra data can be supplied in a third parameter */
    HashReturn (*init)  (hashState *, int, ... );
    /* the update function  */
    HashReturn (*update) (hashState, const BitSequence *, DataLength);
    /* the final function   */
    HashReturn (*final) (hashState, BitSequence *);
    /* the all-in-one function */
    HashReturn (*hash) (int, const BitSequence *, DataLength, BitSequence *);
    /* the free function    */
    void (*free)    (hashState);
    /* hash into byte array */
    HashReturn (*hashtobyte)(hashState, BYTE *out);
    /* the read file function   */
    HashReturn (*file)  (hashState, FILE *in);
    /* pointer to an array of test vectors */
    TestVector      *testvector;
    /* pointer to array of bitwise test vectors */
    TestVectorBitwise   *testvector_b;
    /* ptr to HMAC test vectors */
    HMACTestVector  *hmactestvector;
    /* the length of the produced hash output (bytes)   */
    int  hash_length;
    /* length of input buffer in bytes */
    int  inputbuffer_length;
    /* bibliographic reference of algorithm */
    char *reference;
    /* 1 if bitwise ops are supported */
    /* 2 if bitwise ops are supported, but no published test vectors avail. */
    int  bitwise_support;
    // int  hmac_OK;    // cleared for sha3-*, shake*
    int  mac_OK;    // 1 for HMAC, 2 for KMAC (bitmask), 4 for KMACxxx
    int  base64_OK; // set if base64 output of hash is OK
    int  xof_OK_defaultLength;  /* set to default length if variable 
                                 * output length is OK (XOFSHAKExxx) */
#endif
void print_kmacxxxtestvector(KMACxxxTestVector *t);

int TestSuite (struct use_algo *ua, struct test_results *t)	/* returns 0 for success */
{
	TestVector	*testvector;

	unsigned char out [HASHBUFSIZ];	/* long enough to keep the hash */
	int	i, j;
	int	passed = 0, failed = 0, totalfailed = 0;;

	struct hash_algo *a = ua->a;

	if (!a) return 1;
	if (!a->testvector) return 0;

	t->name = a->name;
	t->test_type = "Hash tests";
	t->tests_ok = 0;
	t->tests_failed = 0;
	
	testvector = a->testvector;
	printf ("%s Test Suite:\n", a->name);

    for (i=0; testvector->string && testvector->repeat > 0; i++, testvector++) {
        void *context = NULL;
		HashReturn retval = (*a->init)(&context, a->hash_length<<3, &ua->extra);
		if (retval) hash_error (retval);
		if (!context) {
			fprintf (stderr, "Testsuite: got no context for %s Test %d\n",
				a->name, i);
			exit (1);
		}
		int	 length = strlen (testvector->string);
		/* testvector in hex notation */
		if (testvector->keytype == IS_HEXSTRING) {
			/* convert testvector into a byte array, may have blanks etc */
			/* allocate temp buffer to hold string without illegal chars */
			char *temp = (char *) malloc (length);
			int k, ktemp;
			for (k=0, ktemp=0; k<length; k++) {
				/* skip non-hex characters */
				int c = testvector->string[k];
				if (!isxdigit(c)) continue;	/* skip non-hex, like blanks */
				temp[ktemp++] = c;
			}
			/* temp is now pure hex, check that length is even */
			if (length%2) {
				fprintf (stderr, "%s (\"%s\") is hex string, but length %d is not even.\n",
					a->name, testvector->string, ktemp);
				exit (1);
			}
			/* allocate hexbuf buffer to hold BYTE values */
			BYTE *hexbuf = (BYTE *) malloc (ktemp/2);
			int khex = 0;
			/* now convert temp into BYTE array hexbuf */
			for (k=0; k<ktemp; k +=2) {
				unsigned int c;
				sscanf (temp+k, "%2x", &c);
				hexbuf[khex++] = c;
			}
			free (temp);
		
			/* digest the testvector 1 or repeat times	*/
			for (j=0; j<testvector->repeat; j++)
				(*a->update)(context, hexbuf, khex<<3);
			free (hexbuf);
			printf ("%s (\"hexstring=%s, bits=%d\")", a->name, 
				testvector->string, khex<<3);
		}
		else {	/* normal string */
			for (j=0; j<testvector->repeat; j++)
				(*a->update)(context, (BYTE *) testvector->string, length<<3);
			printf ("%s (\"%s\") ", a->name, testvector->string);
        }
        // final processing for both hex and ASCII
		(*a->final)(context, NULL);
		// make sure out is large enough
		assert (sizeof(out) >= a->hash_length);
       	(*a->hashtobyte) (context, out);
		//printf ("%s (\"%s\") ", a->name, testvector->string);
		
		if (testvector->repeat>1)
			printf ("(repeated %d times)", testvector->repeat);
		printf (" = ");
		for (j=0; j<a->hash_length; j++)
			printf ("%2.2x", out[j]);
		if (verify_hash (out, testvector->digest, a->hash_length)) {
			passed++; printf ("  OK\n");
		} else {
			failed++; printf ("  failed.\n");
			printf ("The description %s expects the result: %s\n",
				a->reference, testvector->digest);
		}
        (*a->free) (context);
    }
	totalfailed = failed;
	printf ("Tests passed: %d, tests failed: %d.\n", passed, failed);
	t->tests_ok += passed;
	t->tests_failed += failed;

	return (totalfailed);
}
	
int TestSuiteBitwise (struct use_algo *ua, struct test_results *t)	
	/* returns 0 for success */
	/* bitwise test vectors if available */
{
	unsigned char out [HASHBUFSIZ];	/* long enough to keep the hash */
	int	i;
	int	passed = 0, failed = 0, totalfailed = 0;

	struct hash_algo *a = ua->a;

	if (!a) return 1;
	if (!a->testvector_b) return (0);

	t->name = a->name;
	t->test_type = "Bitwise hash tests";
	t->tests_ok = 0;
	t->tests_failed = 0;
	
	TestVectorBitwise	*tvb = a->testvector_b;
	printf ("\n%s Bitwise Test Suite:\n", a->name);
	passed = failed = 0;

    for (i=0; tvb->type; i++, tvb++) {
        void *context = NULL;
		HashReturn retval = (*a->init)(&context, a->hash_length<<3, &ua->extra);
		if (retval) hash_error (retval);
		if (!context) {
			fprintf (stderr, "Bitwise Testsuite: got no context for %s Test %d\n",
				a->name, i);
			exit (1);
		}
		/* Gillogly input is only tested for SHA1
	 	 * however, it should work on SHA224, SHA256, SHA384, SHA512
		 * but has not been tested (no test vectors available)
		 */
		if (tvb->type == IS_BITSTRING) {
			do_BitStringGillogly (a, context, tvb->bitstring);

			// make sure out is large enough
			assert (sizeof(out) >= a->hash_length);
			(*a->hashtobyte) (context, (unsigned char *) out);
			printf ("%s (\"bitstring=%s\") = ", a->name, tvb->bitstring);
			for (i=0; i<a->hash_length; i++)
	   				printf ("%2.2x", out[i]&0xff);
			if (verify_hash (out, tvb->digest, a->hash_length)) {
				passed++; printf ("  OK\n");
			} else {
				failed++; printf ("  failed.\n");
				printf ("The description %s expects the result: %s\n",
					tvb->bitreference, tvb->digest);
			}
			(*a->free) (context);
		}
		if (tvb->type == IS_HEXSTRING) {
			do_HexString (a, context, tvb->hexstring, (uint64) tvb->bitsize, 1);

			// make sure out is large enough
			assert (sizeof(out) >= a->hash_length);
			(*a->hashtobyte) (context, (unsigned char *) out);
			printf ("%s (\"hexstring=%s, bits=%d\") = ", a->name, 
					tvb->hexstring, tvb->bitsize);
			for (i=0; i<a->hash_length; i++)
	   				printf ("%2.2x", out[i]&0xff);
			if (verify_hash (out, tvb->digest, a->hash_length)) {
				passed++; printf ("  OK\n");
			} else {
				failed++; printf ("  failed.\n");
				printf ("The description %s expects the result: %s\n",
					tvb->bitreference, tvb->digest);
			}
			(*a->free) (context);
		}
    }
	totalfailed += failed;
	printf ("Tests passed: %d, tests failed: %d.\n", passed, failed);
	t->tests_ok += passed;
	t->tests_failed += failed;
	return (totalfailed);
}

// int TestSuiteHMAC (struct hash_algo *a, struct test_results *t)	
int TestSuiteHMAC (struct use_algo *ua, struct test_results *t)	
		/* returns 0 for success */
{
	HMACTestVector	*testvector;

	unsigned char out [HASHBUFSIZ];	/* long enough to keep the hash */
	int	i, j;
	int	passed = 0, failed = 0, totalfailed = 0;;
	struct	use_algo	use;

	struct hash_algo *a = ua->a;
	
	if (!a) return 1;
	if (!a->hmactestvector) return (0);
	if (!hmac_OK(a->mac_OK)) return (0);

	t->name = a->name;
	t->test_type = "HMAC tests";
	t->tests_ok = 0;
	t->tests_failed = 0;

	testvector = a->hmactestvector;
	printf ("\n%s HMAC Test Suite:\n", a->name);

    for (i=0; testvector->keytype && testvector->stringtype; i++, testvector++) {
        void *context;
		HashReturn retval = (*a->init)(&context, a->hash_length<<3, &ua->extra);
		if (retval) hash_error (retval);
		if (!context) {
			fprintf (stderr, "TestsuiteHMAC: got no context for %s Test %d\n",
				a->name, i);
			exit (1);
		}
		keystring = testvector->key;
		hexkeystring = testvector->hexkey;
		memset (&use, 0, sizeof (use));
		use.a = a;
		use.hmacflag = 1;
		HMAC_INIT ((&use), testvector->keytype == IS_ASCIISTRING, hmacprintflag);
		HMAC_UPDATE ((&use), context);
		if (testvector->stringtype == IS_ASCIISTRING) {
			(*a->update)(context, (BYTE *) testvector->string, 
						strlen (testvector->string)<<3);
			(*a->final)(context, NULL);
		}
		else {
			// do_HexString (a, context, testvector->hexstring, 0LL);
			do_HexString (a, context, testvector->hexstring, strlen(testvector->hexstring)*4, 0);   // no bitwise for HMAC
		}
		// make sure out is large enough
		assert (sizeof(out) >= a->hash_length);
        (*a->hashtobyte) (context, out);
        (*a->free) (context);
		HMAC_FINAL ((&use), out);
		printf ("hmac-%s (\"%s\") ", a->name, 
			testvector->stringtype == IS_ASCIISTRING ? testvector->string : 
			testvector->hexstring);
		printf (" = ");
		for (j=0; j<a->hash_length; j++)
			printf ("%2.2x", out[j]);
		if (verify_hash (out, testvector->digest, a->hash_length)) {
			passed++; printf ("  OK\n\n");
		} else {
			failed++; printf ("  failed.\n");
			printf ("The description %s expects the result: %s\n\n",
				testvector->hmacreference, testvector->digest);
		}
    }
	totalfailed = failed;
	printf ("Tests passed: %d, tests failed: %d.\n", passed, failed);
	t->tests_ok += passed;
	t->tests_failed += failed;

	return (totalfailed);
}
#if 0
typedef struct {
    char    *key;
    // char *hexkey;
    int     keytype;    // IS_HEXSTRING, IS_ASCIISTRING
    int     keylen;     // key length in bits if specified
    char    *string;    // also called msg, message
    // char *hexstring;
    int     stringtype; // IS_HEXSTRING, IS_ASCIISTRING
    int     stringlen;  // string length in bits if specified
    char    *customstring;
    int     customstringtype;   // NEW!! IS_HEXSTRING, IS_ASCIISTRING
    BitLength customBitLen;
    char    *digest;    // also called mac
    int     digestlen;
    char    *kmacreference;
} KMACxxxTestVector;

#endif

#if 0
/* we need this because myKMAC128_hash etc need these parameters, and we cannot
 * feed them otherwise, without breaking the pattern
 */
extern char    *keystring;      // ASCII key string (for xMAC)
extern char    *hexkeystring;   // key string in hex (for xMAC)
extern int     keylength;       // length of hexkeystring in bits for KMAC

extern int     customflag;
extern char    *customstring;
extern BitLength customBitLen;
#endif


int TestSuiteKMACxxx (struct use_algo *ua, struct test_results *t)	
		/* returns 0 for success */
{
	KMACxxxTestVector	*testvector;

	unsigned char out [HASHBUFSIZ];	/* long enough to keep the hash */
	int	i, j;
	int	passed = 0, failed = 0, totalfailed = 0;
	struct	use_algo	use;

	struct hash_algo *a = ua->a;
	
	if (!a) return 1;
	if (!a->hmactestvector) return (0);
	if (!kmacxxx_OK(a->mac_OK)) return (0);

	t->name = a->name;
	t->test_type = "KMAC tests";
	t->tests_ok = 0;
	t->tests_failed = 0;

	testvector = (KMACxxxTestVector *) a->hmactestvector;
	printf ("\n%s KMAC Test Suite:\n", a->name);
    

    for (i=0; testvector->keytype && testvector->stringtype; i++, testvector++){
        // print_kmacxxxtestvector(testvector);
        // printf ("\n");

		char *keystring = testvector->key;
        int  keytype = testvector->keytype;
        int  keylen  = testvector->keylen;
        // printf ("\nTestVector Data:\n");
        // printf ("keystring = %s, keytype %d, keylen %d\n", keystring, keytype, keylen);

        char *mystring = testvector->string;
        int  mystringtype = testvector->stringtype;
        int  mystringlen  = testvector->stringlen;
        // printf ("mystring = %s, mystringtype %d, mystringlen %d\n", mystring, mystringtype, mystringlen);

        char *mycustomstring = testvector->customstring;
        int  customstringtype = testvector->customstringtype;
        BitLength mycustomBitLen = testvector->customBitLen;
        // printf ("mycustomstring = %s, customstringtype %d, mycustomBitLen %ld\n", mycustomstring, customstringtype, mycustomBitLen);

        char *digest = testvector->digest;
        int  digestlen  = testvector->digestlen;

        char *kmacreference = testvector->kmacreference;
        // printf ("digest = %s, digestlen = %d, kmacreference %s\n\n", digest, digestlen, kmacreference);

        /* convert hex strings into BYTE arrays */
        // key
        if (keytype == IS_HEXSTRING) {
           int outlen = strlen(keystring);
            BYTE *out = (BYTE *) malloc (outlen+1);
            // copy hexkeystring into out
            int nibcount = copyhextobin (out, outlen,
                                keystring, strlen(keystring));
            if (nibcount&0x01) {
                fprintf (stderr, "copyhextobin: %s has odd nibbles %d, stop\n",
                    hexkeystring, nibcount);
                exit (1);
            }
            ua->extra.key = out;
            if (keylen > 0) ua->extra.KeyBitLen = keylen;
            else ua->extra.KeyBitLen = nibcount*4;
        }
        else if (keytype == IS_ASCIISTRING) {
            ua->extra.key = keystring;
            if (keylen > 0) ua->extra.KeyBitLen = keylen;
            else ua->extra.KeyBitLen = strlen(keystring)<<3;
        }
        else {
            fprintf (stderr, "TestSuiteKMACxxx key: only hex and ASCII.");
            exit (1);
        }

        // string
        if (mystringtype == IS_HEXSTRING) {
            int outlen = strlen(mystring);
            BYTE *out = (BYTE *) malloc (outlen+1);
            // copy hexkeystring into out
            int nibcount = copyhextobin (out, outlen,
                                mystring, strlen(mystring));
            if (nibcount&0x01) {
                fprintf (stderr, "copyhextobin: %s has odd nibbles %d, stop\n",
                    keystring, nibcount);
                exit (1);
            }
            mystring = out;
            if (mystringlen == 0) mystringlen = nibcount*4;
        }
        else {
            mystringlen = strlen(mystring)<<3;
        }
    
        // customization string
        if (customstringtype == IS_HEXSTRING) {
            int outlen = strlen(mycustomstring);
            BYTE *out = (BYTE *) malloc (outlen+1);
            // copy hexkeystring into out
            int nibcount = copyhextobin (out, outlen,
                                mycustomstring, strlen(mycustomstring));
            if (nibcount&0x01) {
                fprintf (stderr, "copyhextobin: %s has odd nibbles %d, stop\n",
                    mycustomstring, nibcount);
                exit (1);
            }
            mycustomstring = out;
            if (mycustomBitLen == 0) mycustomBitLen = nibcount*4;
        }
        else {
            mycustomBitLen = strlen(mycustomstring)<<3;
        }
        ua->extra.customize = mycustomstring;
        ua->extra.customizeBitLen = mycustomBitLen;
        // print_EXTRA (&ua->extra);
        void *context = NULL;
		HashReturn retval = (*a->init)(&context, a->hash_length<<3, &ua->extra);
		if (retval) hash_error (retval);
		if (!context) {
			fprintf (stderr, "TestsuiteKMACxxx: got no context for %s Test %d\n",
				a->name, i);
			exit (1);
		}
        if (mystringtype == IS_HEXSTRING) {
#if 0
        printf ("mystring as hex, %d bits  = ", mystringlen);
            for (int i=0; i<(mystringlen+7)/8; i++)
                  printf ("%2.2x ", mystring[i]& 0xff);
                printf ("\n"); 
        else
            printf ("mystring as ascii = %s, mystringlen %d\n", mystring, mystringlen);
#endif
        }
        (*a->update)(context, mystring, mystringlen);
    	(*a->final)(context, NULL);
		
		// make sure out is large enough
		assert (sizeof(out) >= a->hash_length);
        (*a->hashtobyte) (context, out);
        (*a->free) (context);
		printf ("%s (\"%s\") ", a->name, testvector->string);
		printf (" = ");
		for (j=0; j<a->hash_length; j++)
			printf ("%2.2x", out[j]);
		if (verify_hash (out, testvector->digest, digestlen? digestlen : a->hash_length)) {
			passed++; printf ("  OK\n\n");
		} else {
			failed++; printf ("  failed.\n");
			printf ("The description %s expects the result: %s\n\n",
				testvector->kmacreference, testvector->digest);
		}
    }
	totalfailed = failed;
	printf ("Tests passed: %d, tests failed: %d.\n", passed, failed);
	t->tests_ok += passed;
	t->tests_failed += failed;

	return (totalfailed);
}

void print_kmacxxxtestvector(KMACxxxTestVector *t)
{
    printf ("KMACxxxTestVector\n");
    printf ("key    : %s\n", t->key);
    printf ("keytype: %d\n", t->keytype);
    printf ("keylen : %d\n", t->keylen);
    printf ("string : %s\n", t->string);
    printf ("stringtype: %d\n", t->stringtype);
    printf ("stringlen : %d\n", t->stringlen);
    printf ("customstring = %s\n", t->customstring);
    printf ("customstringtype: %d\n", t->customstringtype);
    printf ("customBitLen = %ld\n", t->customBitLen);
    printf ("digest = %s\n", t->digest);
    printf ("digestlen : %d\n", t->digestlen);
    printf ("kmacreference = %s\n", t->kmacreference);
}
